/*
 * LCD_Config v4.0p1
 * -----------------
 *
 * last change: 12.08.2011
 *
 * Allows configuration of the MultiWii software via a serial connection.
 * Requires a LCD display and at least four buttons (left, right, up, down)
 * or an analog joystick. The SELECT button is for both configurations optional.
 *
 * Options:
 *   Logging functionality (requires a SD card), output to binary or LogView OpenFormat.
 *   Status LEDs (three LEDs, two are sufficient without logging)
 *   Setting of LCD backlight intensity
 *   Use of ShiftRegLCD lib instead of LiquidChrystal lib
 *   Use of EagleTree PowerPanel
 *   Use of Sparkfun Serial LCD
 *   Emulation of a Textstar LCD to configure the copter via the LCD_CONF functionality
 *   Hardware buttons instead of buttons connected via cascading resistors
 *   Analog joystick instead of buttons
 *   Graphic output (simple character graphic for angles and servo signals)
 *   MinMax mode (store und display minimum/maximum values of certain fields)
 *   [Test] (for standalone testing without a MWC board connected)
 *   [CHECK_INPUT] (presents the state of the input pins - press RIGHT or SELECT in start state for >1 sec)
 *   [DUMP_BUFFER] (dumps the bytes in the serial buffer to the LCD (in hexadecimal) when an invalid
 *   frame is received (useful for debugging))
 *   [APC220] (disables a APC220 wireless module during playback of logfile - rather specific option,
 *   don't use it unless you have an APC module) 
 *
 * IMPORTANT: Due to a bug in the way Arduino tread #include commands, all #includes are now in
 * a separate block at the beginning of the Sketch. When configuring the Sketch to your hardware, 
 * you have to select your set of options and the libraries these options require separately.
 * Other way round, you have to comment/uncomment lines twice. Annoying, yes, but it saves
 * memory since the unused libraries are linked to the executable otherwise.
 *
 * ----------
 *
 * Graphic Mode is toggled either by pressing the SELECT key or by pressing the RIGHT key longer
 * than 1 sec. Min/Max Mode is toggled by pressing the RIGHT key shorter than 1 sec.
 *
 * ----------
 *
 * For infos on LogView fileformat, see www.logview.info.
 * To import logfiles into LogView, you need to copy the attached .ini and .jpg File
 * to the directory <username>/<application data>/LogView/OpenFormat, select the
 * device LCD_CONFIG, and click on "import device file".
 * Data is formatted to three LogView channels: the first channel contains sensor data,
 * the second servo and motor PWM signals, the third the receiver channels. 
 *
 * ----------
 *
 * Authors: yacco, 
 *          calli        (HWKEYS, ShiftRegLCD), 
 *          Danal Estes  (ETPP [Eagle Tree Power Panel] support)
 *          dynai        (early version of EMULATE_TEXTSTAR)
 *
 * changes to 1.25: 
 *   display values of gyro and acc changed to signed
 *   show minimum/maximum values of sensor data and servo signals
 * changes to 1.28:
 *   wait state after startup, no serial communication during initialization of the MWC
 *   removed min/max values
 *   added LED status signals
 * changes to 1.30:
 *   bugfix in playback routine, added timeout in playback
 * changes to 1.31:
 *   minor bugfixes, added setting of backlight
 * changes to 1.32:
 *   removed KEY_SELECT column in next_state, backlight and waitForInit optional
 * changes to 1.33:
 *   check trailing 'A' char in serial communication to detect frame errors, 
 *   optimization of format routines (merged two routines, invalid values are displayed as ###)
 *   bugfix in logging (logMode=-1 was stored in flash, so logging was permanently disabled)
 *   KEY_LEFT cancels reading when no serial connection is present (important to reset wrong settings!) 
 * changes to 1.5:
 *   reimplemented Min/Max-Mode (found it useful again), 
 *   fixed compiling problems for certain combinations of options 
 * changes to 1.6:
 *   adaptation to MWC version 1.preter6 (new config "Flying Wing", new data format)
 *   changed status display (state 0 - Config)
 *   bug 'logMode=-1' finally solved
 *   Version has been set to 2.0 to disambiguate between MWC version and LCD_Conf version.
 * changes to 2.0 beta: 
 *   swapped motors 4 and 5 in 'Esc Out' for versions 1.5 and 1.6prebis
 *   changed behaviour of UP and DOWN for consistency (UP decrements, DOWN increments values)
 *   added version V1.6 
 *   output logfiles in LogView format [OPTIONAL] (file numbers starting from 500, extension ".lvw")
 *   added JOYSTICK option to replace input buttons by an analog joystick [experimental]
 *   display 'invalid' values in graphics
 *   heavy optimizations (!) due to space restrictions
 *   removed WAIT_FOR_INIT option, changed start state to 31 instead 
 * changes to 2.0:
 *   bugfix in deserialize() with v1.6
 * changes to 2.01:
 *   Joystick handling was improved (calibration, use of pullups, can be used in parallel to buttons)
 *   Added 'Play LogFile' state
 *   Easter egg function to display internal configuration :)
 *   BUGFIX: RC Rate multiplied by factor 2 for MWC Versions >v1.5, minimum of 20% for RC Rate (0% makes no sense)
 *   Added random offsets in TEST mode
 * changes to 2.1:
 *   Cancel in Logfile Playback (press LEFT_KEY)
 *   Settings of "Baud Rate", serialRead() has been adapted to work with different COM speeds
 *   Greater flexibility in changing the Settings
 *   (Log) Delay Time was moved to Settings state and is used as the time interval of polling the MWC board
 * changes to 2.2:
 *   Changed internal representation of state machine and strings
 *   Support for v1.7pre (I-/P-Level instead of AutoLevelStrength, trigger calibration, 8 instead of 5 receiver
 *   channels, configuration of activation modes for Aux1/Aux2, ...)
 *   Software Version is now stored in Logfile
 *   APC220 Option (disables APC220 module via pin D2 during playback of logfiles, experimental)
 * changes to 2.3:
 *   Swapped Angle and Baro/Comp/Cycle states
 *   Added 'Config' parameter to LogFiles, show Config and Version in selectLogFile
 *   Use of blinking Cursor instead of Arrows in Enter mode
 *   LogView mode: corrected order of receiver channels
 * changes to 2.4:
 *   "Roll/Pitch" state - simultaneous setting of Roll and Pitch PID values
 * changes to 2.5:
 *   Changed representation of configIds
 *   'pos' parameter in lcd output instead of (col, row), saved ~250 bytes of flash
 *   Support for v1.7prebis (replaced v1.7pre)
 *   CHECK_INPUT option to visualize the joystick input (hidden state, press RIGHT in state 0 for >1 sec)
 *   BATTERY_LEVEL option (experimental), visualisation on LCD and via Status LEDS, 4 warn levels can be defined
 *   Optimized presentation of ESC/Servo signals (saved ~750 bytes)
 * changes to 2.6:
 *   Added consistency check in serialRead, removed strange behaviour after framelength errors.
 *   Improved error handling when communication fails (more detailed description than "Inv. FrameLength")
 *   Show MWC version in start state
 *   Show current values in "Settings" state, added blinking cursor when entering values
 *   Number of Profiles can be configured (constant NUM_PROFILES)
 * changes to 2.7:
 *   Removed BATTERY_LEVEL problem in conjunction with APC220
 *   Flashing in battery status display to indicate warn levels
 *   Bugfix in setting of delay time
 * changes to 2.8:
 *   KEY_PIN can be undefined if no cascading Buttons are used
 *   Changed default values for Version (highest version) and Baudrate (115200)
 *   Added Option "Dump Buffer" to show the contents of the data buffer when an invalid frame is received (for debugging)
 *   DEFAULT_BAUDRATE can be configured, USE_DEFAULT_BAUDRATE ignores stored Baudrate params (Baudrate can still be adjusted manually)
 *   Bugfix in state "Enter R/C Rate / Expo" (no action on RIGHT)
 *   Bugfix concerning Battery level (removed leading extra byte)
 * changes to 2.9:
 *   Removed JS_INV_SELECT_KEY (Joystick Select key has to be connected to GND when closed).
 *   Bugfix in state "Enter Level PID" (no action on RIGHT)
 * changes to 3.0:
 *   Eagle Tree Power Panel (Option LCD_ETPP) support (setting of backlight not supported)
 *   Sparkfun Serial LCD (Option SERIAL_LCD) support (graphic mode not supported)
 * changes to 3.1:
 *   changed Serial LCD from SoftwareSerial lib to NewSoftSerial lib (see http://arduiniana.org/libraries/NewSoftSerial/)
 *   added EMULATE_SERIAL_LCD option to use the LCD_Config as a Serial LCD (option LCD_CONF has to be enabled in MWC software).
 *   - To enter the emulation state press KEY_RIGHT during the initial 'Reading' state. 
 *   - This feature allows you to use the LCD_Config with MWC versions not supported by LCD_Config (i.e., the developer version).
 *   bugfix in restoreParams()
 *   in "Reading..." state, the characters '<' and '>' indicate which keys are currently selectable. ('<' to cancel, '>' to enter
 *   the fabulous 'Serial LCD Emulation' mode.)
 *   extracted option section to separate Header file, the user should ONLY edit this file and the comments in the #include lines below
 * changes to 3.2
 *   replaced EMULATE_SERIAL_LCD by EMULATE_TEXTSTAR
 *   added support for MWC Version 1.8
 *   therefor, removed support for MWC Version < 1.7 (!)
 *   removed BATTERY_LEVEL option (now mandatory), added support for PowerMeter
 * changes to 4.0pre:
 *   formatBar() for Textstar Emulation Mode
 *   Bugfix in resetProfile()
 *   Bugfix in playbackLog()
 *   Optimized write access to EEPROM
 *
 * overall footprint (on an Arduino Uno, using the LiquidCrystal lib and cascading buttons): 
 *  ~28228 bytes with all options enabled (the maximum image size uploadable via bootloader on a AtMega328 is ~28670 bytes)
 *  ~16052 bytes without logging, status leds and APC220 (default config)
 *  ~12220 bytes with all options disabled
 */

// ----- begin of #includes -----

// include Constants/Macros for Options 
#include "LCD_Config.h"

// for persistence of SW Version, Logging Mode, DelayTime and Profiles (mandatory)
#include <EEPROM.h>

// defines the PROGMEM macro to move constant variables from SRAM to flash memory (mandatory)
#include <avr/pgmspace.h> 

// comment this line when NOT using a regular HD44780
#include <LiquidCrystal.h>

// uncomment this line when using ShiftReg LCD (Option SRLCD)
//#include <ShiftRegLCD.h>      

// uncomment this line when using a Serial LCD (Option SERIAL_LCD)
//#include <NewSoftSerial.h>

// uncomment this line when LOGGING is enabled
#include <SD.h>

// Workaround for "insertion of forward declarations in excluded blocks" problem
// Don't move this line around.
void setup();     

// ----- end of #includes -----

/* ----- States: ----- */

#ifdef LOGGING
  #define NUM_STATES  41
#else
  #define NUM_STATES  40
#endif

// Number of Configurations: (Unknown, Tri, Quad+, QuadX, Bi, Gimbal, Y6, Hex, Flying Wing, Y4, HexX, Oct)
#define NUM_CONFIGS  12

// the GUI_SERIAL_SPEED is used for logfile playback only
#define GUI_SERIAL_COM_SPEED 115200

/* ----- LCD Configurations: ----- */

// HD44780 LCD via the ShiftReg Library
#if defined(SRLCD)
  ShiftRegLCD                lcd(3, 2, 4, 2); // (Datapin, Clockpin, Enablepin or TWO_WIRE [, Lines [, Font]]])
  #define INIT_LCD                     // no begin in ShiftRegLCD...

// Eagle Tree Power Panel
#elif defined(LCD_ETPP)
  #define INIT_LCD i2c_init();  i2c_ETPP_init();
  #undef BACKLIGHT                     // backlight not supported

// Sparkfun Serial LCD
#elif defined(SERIAL_LCD)
  #define SERIAL_LCD_TX      2         // emulates serial port with TX on D2
  #define LCD_DELAY          4         // conservative, 2 actually works
  #define INIT_LCD           pinMode(SERIAL_LCD_TX, OUTPUT); lcd.begin(9600);

  #undef GRAPHIC                       // custom characters not supported

  NewSoftSerial             lcd = NewSoftSerial(0, SERIAL_LCD_TX);
  void lcdDelay()           {delay(LCD_DELAY);}  // delay for transmission

// Regular HD44780 via Liquid Crystal Library (e.g., DFRobot Keypad Shield)
#else
  LiquidCrystal             lcd(8, 9, 4, 5, 6, 7); // rs, enable, d4, d5, d6, d7, 
  #define INIT_LCD          lcd.begin(16, 2); 
#endif

#if defined(LCD_ETPP)
  #define LEFT_ARROW  "\074"
  #define RIGHT_ARROW "\076"
  #define DOT         '.'
#else
  #define LEFT_ARROW  "\177"
  #define RIGHT_ARROW "\176"
  #define DOT         165
#endif  

/* ----- LCD Backlight: ----- */

static uint8_t backlight;              // backlight intensity (0..255)

#ifdef BACKLIGHT
  #define BL_MSG             "Backlight"
  #define NUM_SETTINGS       3
  #define INIT_BACKLIGHT     ee_ptr = 5; backlight = 255-readEEPROM(); setBacklight();

  #ifdef SERIAL_LCD
    void setBacklight(uint8_t arg) {uint8_t bl = 157 - max(arg/5, 29); lcdCmd(bl);}      
  #else
    void setBacklight(uint8_t arg) {analogWrite(BACKLIGHT_PIN, arg);}
  #endif
#else
  #define BL_MSG
  #define NUM_SETTINGS       2
  #define INIT_BACKLIGHT
  void setBacklight(uint8_t arg)   {}
#endif

void setBacklight()          {setBacklight(backlight);}



/* ----- String Literals and Constants: ----- */

#define VERSION "4.01"

// cryptic representation of selected compiler options.
// can be visualized at run-time by pressing LEFT in start state for >1 sec.
// useful when testing different sets of options
PROGMEM prog_char metaData[] = VERSION ":" 
#ifdef TEST
  "T"
#endif
#ifdef SRLCD
  "S"
#endif
#ifdef LCD_ETPP
  "P"
#endif
#ifdef SERIAL_LCD
  "C"
#endif
#ifdef EMULATE_TEXTSTAR
  "X"
#endif
#ifdef KEYPIN
  "K"
#elsifdef HWKEYS
  "H"
#endif
#ifdef JOYSTICK
  "J"
#endif
#ifdef GRAPHIC
  "G"
#endif
#ifdef MINMAX
  "M"
#endif
#ifdef LOGGING
  "L"
  #ifdef LOGVIEW
  "v"
  #else
  "-"
  #endif
#endif
#ifdef BACKLIGHT
  "B"
#endif
#ifdef STATUS_LEDS
  "$"
#endif
  ":"
;

// ----------------

PROGMEM prog_char intro         [] = " LCD Config " VERSION;
PROGMEM prog_char emptyLine     [] = "\020";
#ifdef EMULATE_TEXTSTAR
  PROGMEM prog_char reading     [] = "Reading:\006" LEFT_ARROW RIGHT_ARROW;  // '<>' indicates KEY_LEFT and KEY_RIGHT are active
  PROGMEM prog_char emuLcdMsg   [] = "TEXTSTAR Mode\004";
#else
  PROGMEM prog_char reading     [] = "Reading:\007" LEFT_ARROW;              // '<' indicates KEY_LEFT is active
#endif
PROGMEM prog_char writeMsg      [] = " Wrote\004Bytes ";
PROGMEM prog_char cancelRead    [] = "Reading canceled";
PROGMEM prog_char wrongMsg      [] = "Wrong\011";
PROGMEM prog_char timeoutError  [] = " Timeout Error\002";
PROGMEM prog_char calibMsgAcc   [] = " Calibrate Acc? ";
PROGMEM prog_char calibMsgMag   [] = " Calibrate Mag? ";
PROGMEM prog_char yesNoMsg      [] = " " LEFT_ARROW ": no, " RIGHT_ARROW ": yes\002";
PROGMEM prog_char delaySuff     [] = "\0100 ms";
PROGMEM prog_char offlineMsg    [] = "Offline";
PROGMEM prog_char haltMsg       [] = " Program halted ";

PROGMEM prog_char frameError    [][ 7] = { "Length", "Prefix", "Suffix", "Config", "Vers.", };
PROGMEM prog_char profStates    [][ 6] = { "Load", "Save", "Clear", "Reset" }; 
PROGMEM prog_char auxStates     [][ 7] = { "Level", "Baro", "Magnet", "C-Stab", "C-Trig", "Arming", }; 
PROGMEM prog_char setMsgs       [][11] = { "Delay Time", "Baud Rate", BL_MSG };

#ifdef LOGGING
  PROGMEM prog_char logMsg      [] = "Logging...\006";
  PROGMEM prog_char noCard      [] = "No SD Card found";
  PROGMEM prog_char fileError   [] = "Open File failed";
  PROGMEM prog_char deleteFailed[] = "Delete failed\003";
  PROGMEM prog_char haltReqMsg  [] = " Halt Program?\002";
  PROGMEM prog_char logEnbMsg   [] = "Logging enabled ";
  PROGMEM prog_char stopLogMsg  [] = " Stop Logging?\002";
  PROGMEM prog_char fileMsg     [] = "File: ";

  PROGMEM prog_char logStates   [][7] = { "Start", "Play" , "Delete" };

  #ifdef LOGVIEW
    PROGMEM prog_char formatMsg [] = "LogView Format?";
  #endif
#endif

/* ----- */

#ifdef MINMAX
  PROGMEM prog_char minMaxDsc   [] = "Act\0Min\0Max";
#endif

/* ----- */

#ifdef TEST
  #define INIT_TEST lcdPrint(12 + 16, "Test"); randomSeed(4711); config = 1;
#else
  #define INIT_TEST
#endif

/* ------ */

#ifdef CREDITS
PROGMEM prog_char credits[] = { 40, 99, 41, 50, 48, 49, 49, 32, 98, 121, 32, 121, 97, 99, 99, 111, 0 };
#endif
int8_t  baudRates        [] = { 1, 2, 4, 8, 12, 16, 32, 48, 96, };
uint8_t defaultProfile   [] = { 18,                     // v1.8
                                40, 30, 17,             // pid roll
                                40, 30, 17,             // pid pitch
                                85,  0,  0,             // pid yaw
                                47,  0,  0,             // pid alt
                                 0,  0,  0,             // pid vel
                                90, 45,                 // pi  level
                                40,                     // p   mag
                                45, 65,  0,  0,  0,     // dual rate, expo, roll/pitch rate, yaw rate, dyn phr pid
                                 0,  0,  0,  0,  0,  0, // activation: level, baro, mag, camStab, camTrig, arming 
                                 0,  0,                 // power trigger (16 bit)
                             };
uint8_t auxMasks         [] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, }; 

// ----------------

PROGMEM prog_char versions[][5] = {
  "1.8 ", 
};

// number of bytes per frame send from the MultiWii board, dependant on version
uint8_t frameLength[] = { 
  116,  // 1.8
};

/* ----- */

// configuration ids
PROGMEM prog_char configIds[] = { 
  "Unknown\0" 
  "Tri\0" 
  "Quad+\0" 
  "QuadX\0" 
  "Bi\0"
  "Gimbal\0" 
  "Y6\0" 
  "Hex\0" 
  "Fl.Wing\0"
  "Y4\0"
  "HexX\0"
  "Oct"
};

uint8_t configOffsets [] = { 0, 8, 12, 18, 24, 27, 34, 37, 41, 49, 52, 57, };
#ifdef LOGGING
  char  shortConfigIds[] = "--TrQ+QxBiGmY6H|FwY4H-Oc"; // used to show the Config of a LogFile
#endif

// Tri:         Left, Right, Rudder, Rear
// Quad+:       Front, Rear, Left, Right
// QuadX:       Front Left, Front Right, Rear Left, Rear Right
// Bi:          Motor 0, Motor 1, Servo 0, Servo 1 
// Gimbal:      Roll, Pitch
// Y6:          Up (Left, Mid/Rear, Right), Down (Left, Mid, Right)
// Hex:         Front (Left, Mid, Right), Rear (Left, Mid, Right)
// Flying Wing: Servo 0-3 
// Y4:          Front Left, Front Right, Rear Up, Rear Down
// HexX:        Left (Front, Mid, Rear), Right (Front, Mid, Rear)
// Oct:         1-8, depending on actual config (OctX8, OctFlatP, OctFlatX)

// first entry in layout struct represents the type: 0 - 2 values, 1 - 4 values, 2 - 6 values, 3 - 8 values
// following entries represent the associated motor/servo. (<20: motors, >=20: servos)

uint8_t esc_layout_0 [] = { 1,  2,  1, 20,  0, };            // Tri
uint8_t esc_layout_1 [] = { 1,  3,  0, 2,   1, };            // Q+
uint8_t esc_layout_2 [] = { 1,  3,  1, 2,   0, };            // Qx
uint8_t esc_layout_3 [] = { 1,  0,  1, 20, 21, };            // Bi

uint8_t esc_layout_4 [] = { 0, };                            // Gimbal - #0 only used once, hard wired
uint8_t esc_layout_5 [] = { 2,  2,  0,  1,  3,  5,  4, };    // Y6
uint8_t esc_layout_6 [] = { 2,  3,  4,  1,  2,  5,  0, };    // Hx
uint8_t esc_layout_7 [] = { 1, 20, 21, 22, 23, };            // FW

uint8_t esc_layout_8 [] = { 1,  3,  1,  0,  2, };            // Y4
uint8_t esc_layout_9 [] = { 2,  3,  5,  2,  1,  4,  0, };    // Hxx
uint8_t esc_layout_10[] = { 3, };                            // Oct, hard wired

uint8_t *esc_layouts [] = { esc_layout_0, esc_layout_1, esc_layout_2, esc_layout_3,
                            esc_layout_4, esc_layout_5, esc_layout_6, esc_layout_7,
                            esc_layout_8, esc_layout_9, esc_layout_10,
                          };

uint8_t esc_pos_4[] = { 3, 12, 3+16, 12+16, };               // LCD positions for 4 values
uint8_t esc_pos_6[] = { 2, 7, 12, 2+16, 7+16, 12+16, };      // LCD positions for 6 values

// ----------------

PROGMEM prog_char rollPitchYaw[] = ": Rol Pit Yaw";    

PROGMEM prog_char msg00[] = "1\002Config\007";
PROGMEM prog_char msg01[] = "2\002P-I-D";
PROGMEM prog_char msg02[] = "3\002Rates";
PROGMEM prog_char msg03[] = "4\002Sensors";
PROGMEM prog_char msg04[] = "5\002Receiver In";
PROGMEM prog_char msg05[] = "6\002ESC/Servo Out";
    
PROGMEM prog_char msg06[] = "Roll";
PROGMEM prog_char msg07[] = "Pitch";
PROGMEM prog_char msg08[] = "Yaw";
PROGMEM prog_char msg09[] = "D/R\003Expo\006";
PROGMEM prog_char msg10[] = "R/P\003Yaw\003Thro";
PROGMEM prog_char msg11[] = "Gyr";
PROGMEM prog_char msg12[] = "Acc";
PROGMEM prog_char msg13[] = "Alt\002Head\002Cycle";
    
PROGMEM prog_char msg14[] = "Roll\002Pitch\002Yaw";
PROGMEM prog_char msg15[] = "Thro\002Aux1\002Aux2";

PROGMEM prog_char msg16[] = "Lf:\006Ri:";
PROGMEM prog_char msg17[] = "Ru:\006Re:";
    
PROGMEM prog_char msg18[] = "Lf:\006Ri:";
PROGMEM prog_char msg19[] = "Fr:\006Re:";

PROGMEM prog_char msg20[] = "FL:\006FR:";
PROGMEM prog_char msg21[] = "RL:\006RR:";

PROGMEM prog_char msg22[] = "M1:\006M2:";
PROGMEM prog_char msg23[] = "S1:\006S2:";

PROGMEM prog_char msg24[] = "Roll.:";
PROGMEM prog_char msg25[] = "Pitch:";

PROGMEM prog_char msg26[] = "Up";
PROGMEM prog_char msg27[] = "Dn";

PROGMEM prog_char msg28[] = "Fr";
PROGMEM prog_char msg29[] = "Re";
    
PROGMEM prog_char msg30[] = "\002Rho\006Phi";
    
PROGMEM prog_char msg31[] = "7\002Profiles";
PROGMEM prog_char msg32[] = "8\002Settings";

PROGMEM prog_char msg33[] = "S1:\006S2:";
PROGMEM prog_char msg34[] = "S3:\006S4:";

PROGMEM prog_char msg35[] = "Cam.: Roll Pitch";
PROGMEM prog_char msg36[] = "Level";

PROGMEM prog_char msg37[] = "Aux: 1-LMH 2-LMH";
PROGMEM prog_char msg38[] = "Roll+Pitch";
PROGMEM prog_char msg39[] = "Lf";
PROGMEM prog_char msg40[] = "Ri";
PROGMEM prog_char msg41[] = "Up:\006Dn:";
PROGMEM prog_char msg42[] = "Mag";
PROGMEM prog_char msg43[] = "Alt";
PROGMEM prog_char msg44[] = "Vel";
PROGMEM prog_char msg45[] = "Batt Power Trigg";

/* ----- */

#define LOG_STATE   (NUM_STATES-1)
#define CONF_OFFSET 23

#ifdef LOGGING
  PROGMEM prog_char lgMsg[] = "9\002Logging";

  #define UP0 LOG_STATE
  #define DN0 LOG_STATE
  #define LOG_MSG lgMsg,
#else
  #define UP0 8
  #define DN0 0
  #define LOG_MSG
#endif

// State messages
PROGMEM prog_char* msgs[] = {
 msg00, msg01, msg02, msg03, msg04, msg05, msg06, msg07, msg08, msg09, 
 msg10, msg11, msg12, msg13, msg14, msg15, msg16, msg17, msg18, msg19, 
 msg20, msg21, msg22, msg23, msg24, msg25, msg26, msg27, msg28, msg29, 
 msg30, msg31, msg32, msg33, msg34, msg35, msg36, msg37, msg38, msg39, 
 msg40, msg41, msg42, msg43, msg44, msg45, LOG_MSG
};

// State Transition Table, values above 80 are pseudo-states
PROGMEM prog_char nextState[NUM_STATES][NUM_KEYS + 1] = {
  // Right Up  Down Left Msg Status   --   Status Flag: -1=No Refreh, -2=Refresh Data, >=0: Msg2
  {  100, UP0,   2,   1,   0,  -1, }, //  0 - Config
  {  100, UP0,   2,  36,   0,  -2, }, //  1 - Acc, Bar, Mag

  {    9,   0,   3,   0,   1,  -1, }, //  2 - PID
  {   14,   2,   4,   0,   2,  -1, }, //  3 - Rates
  {   16,   3,   5,   0,   3,  -1, }, //  4 - Sensors
  {   21,   4,   6,   0,   4,  -1, }, //  5 - Rec In
  {   99,   5,   7,   0,   5,  -1, }, //  6 - ESc Out
  {   94,   6,   8,   0,  31,  -1, }, //  7 - Profile
  {   93,   7, DN0,   0,  32,  -1, }, //  8 - Settings

  {   80,  39,  10,   2,  38,  -1, }, //  9 - PID Roll/Pitch
  {   81,   9,  11,   2,   6,  -1, }, // 10 - PID Roll
  {   82,  10,  12,   2,   7,  -1, }, // 11 - PID Pitch
  {   83,  11,  13,   2,   8,  -1, }, // 12 - PID Yaw
  {   84,  12,  37,   2,  43,  -2, }, // 13 - PID Alt

  {   90,  35,  15,   3,   9,  -1, }, // 14 - DR, Expo
  {   91,  14,  35,   3,  10,  -1, }, // 15 - R/P Rate, Yaw Rate, Throttle PID
 
  {  101,  19,  17,   4,  11,  -2, }, // 16 - Gyro
  {  101,  16,  18,   4,  12,  -2, }, // 17 - Acc
  {  101,  17,  20,   4,  42,  -2, }, // 18 - Mag
  {  101,  20,  16,   4,  30,  -2, }, // 19 - Altitde, Heading, Cycle time
  {  101,  18,  19,   4,  13,  -2, }, // 20 - Angles

  {  101,  23,  22,   5,  14,  -2, }, // 21 - R, P, Y
  {  101,  21,  23,   5,  15,  -2, }, // 22 - Thr, Aux1, Aux2
  {  101,  22,  21,   5,  35,  -2, }, // 23 - Cam Rol, Cam Pit

  {  101,  -1,  -1,   6,  16,  17, }, // 24 - Tri
  {  101,  -1,  -1,   6,  18,  19, }, // 25 - Quad+
  {  101,  -1,  -1,   6,  20,  21, }, // 26 - QuadX
  {  101,  -1,  -1,   6,  22,  23, }, // 27 - Bi
  {  101,  -1,  -1,   6,  24,  25, }, // 28 - Gimbal
  {  101,  -1,  -1,   6,  26,  27, }, // 29 - Y6
  {  101,  -1,  -1,   6,  28,  29, }, // 30 - Hex
  {  101,  -1,  -1,   6,  33,  34, }, // 31 - Flying Wing
  {  101,  -1,  -1,   6,  20,  41, }, // 32 - Y4
  {  101,  -1,  -1,   6,  39,  40, }, // 33 - Hex6X
  {  101,  -1,  -1,   6,  28,  29, }, // 34 - Oct

  {   92,  15,  14,   3,  45,  -2, }, // 35 - Battery Level, PowerMeter
  {   96,  97,  97,   0,  37,  -1, }, // 36 - Activation Modes

  {   85,  13,  38,   2,  44,  -2, }, // 37 - PID Vel
  {   86,  37,  39,   2,  36,  -1, }, // 38 - pLevel, iLevel.
  {   87,  38,   9,   2,  42,  -2, }, // 39 - PID Mag

#ifdef LOGGING
  {   98,   8,   0,   0,  46,  -1, }, // 40 - Logging
#endif
  // Right Up  Down Left Msg Status   --   Status Flag: -1=No Refreh, -2=Refresh Data, >0: Msg2
};

// ----------------

// charset used in profile names
static char charsInName[] = " .01234567890_ABCDEFGHIJKLMNOPQRSTUVWXYZ";
static char profileMsg [] = " #00: 01234567  ";
static char statusMsg  [] = "Nck- Baro- Mag- ";

static char lineBuffer [17];

/* ----- Static Variables: ----- */

static uint16_t  ee_ptr                ;

#ifdef LOGGING
static File      logFile               ;
static uint16_t  lineNo                ; // current record number in logFile
static char      fileName[] = "000.log"; // LogFiles consists of a three digit number (000..999) with extension ".log"
#endif

static boolean   initialized           ;
static boolean   graphMode             ; // graphical output of servo signals
static boolean   offline               ;
static int8_t    state                 ; // current ui state
static int8_t    profile               ; // current profile index
static uint8_t   vers                  ; // MultiWii Software Version
static uint8_t   baudRate              ; // baud rate of serial connection to MWC board
static uint8_t   minMaxMode            ; // show act, min or max values
static int8_t    logMode               ; // logging mode: -1 not available, 0 - disabled, 1 - enabled
static uint8_t   keyPressedTime;       ; // time interval between key down and key up in 50ms steps
static uint8_t   auxState              ;
static uint8_t   delayTime             ; // x10 in milliseconds
static uint16_t  rawKey                ;
static uint16_t  lastReadTime          ;

// MWC data
static uint8_t   actMode[6]            ; // activation modes for level, baro, compas, camStab, camTrig, arming (6 bits, 0-2: Aux1 (Low, Mid, High), 3-5: Aux2 (Low, Mid, High)
static uint8_t   accMode, baroMode, magMode;
static uint8_t   accPres, baroPres, magPres;

static uint16_t  cycleTime             ; // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
static uint8_t   config                ; // Tri, Quad, Hex, etc.
static uint8_t   P8[7], I8[7], D8[7]   ; // 8 bits is much faster and the code is much shorter
static uint8_t   rcRate8               ; // dual rate
static uint8_t   rcExpo8               ; // expo
//static uint8_t   accStrength           ; // AutoLevel Strength 0-250 (vers < v1.7pre)
static uint8_t   rollPitchRate         ; // reduces "P" and "D" of Roll/Pitch depending on the associated stick deviation
static uint8_t   yawRate               ; // reduces "P" and "D" of Yaw depending on Rudder stick deviation
static uint8_t   dynThrPID             ; // reduces "P" and "D" of Roll/Pitch when Throttle stick is above 50%

static uint8_t   vbat                  ; // Battery Level
static uint16_t  powerMeterSum, powerTrigger;
static int16_t   debug       [4]       ;  // multi-purpose debug values

static int16_t   accSmooth   [3] MM_DIM;  // projection of smoothed and normalized gravitation force vector on x/y/z axis, as measured by accelerometer       
static int16_t   gyroData    [3] MM_DIM;  // from WMP+
static int16_t   magADC      [3] MM_DIM;  // magnet sensor data
static int16_t   altitude        MM_DIM;  // from barometer
static int16_t   heading         MM_DIM;  // from compas
static int16_t   angle       [2] MM_DIM;  // absolute angle inclination in deg (when Nunchuk/Acc present)
static int16_t   servo       [4] MM_DIM;  // depends on configuration
static int16_t   motor       [8] MM_DIM;  // depends on configuration 
static int16_t   rcHysteresis[8] MM_DIM;  // receiver input

// ----------------

#ifdef CHECK_INPUT
static char       keySymbols [] = { ' ', 'R', 'U', 'D', 'L', 'S', };
PROGMEM prog_char jsStatusMsg[] = "\002Key\002Hor\002Vert";
#endif  

#ifdef JOYSTICK
static int16_t jsHorCtr, jsVertCtr;
static int16_t jsHorAct, jsVertAct;
    
 #ifdef JS_SELECT_PIN
  #define INIT_JOYSTICK pinIn(JS_SELECT_PIN);pinIn(JS_HOR_PIN);pinIn(JS_VERT_PIN);jsHorCtr=analogRead(JS_HOR_PIN);jsVertCtr=analogRead(JS_VERT_PIN);
 #else
  #define INIT_JOYSTICK                      pinIn(JS_HOR_PIN);pinIn(JS_VERT_PIN);jsHorCtr=analogRead(JS_HOR_PIN);jsVertCtr=analogRead(JS_VERT_PIN);
 #endif
#else
  #define INIT_JOYSTICK
#endif

// ----------------

#ifdef APC220
  #define INIT_APC220                            pinOut(APC220_ENABLE_PIN); enableAPC220(true);
  void enableAPC220(boolean value)               {digitalWrite(APC220_ENABLE_PIN, value);}
#else
  #define INIT_APC220
  #define enableAPC220(a)
#endif

// ----------------

#ifdef STATUS_LEDS
  #define INIT_LEDS                              pinOut(RED_LED); pinOut(YELLOW_LED); pinOut(GREEN_LED);

  #define red(a)                                 digitalWrite(RED_LED,    a)
  #define yellow(a)                              digitalWrite(YELLOW_LED, a)
  #define green(a)                               digitalWrite(GREEN_LED,  a)

  void yellowOn ()                               {digitalWrite(YELLOW_LED, true); }
  void yellowOff()                               {digitalWrite(YELLOW_LED, false);}

//  void red   (boolean a)                       {digitalWrite(RED_LED,    a);}
//  void yellow(boolean a)                       {digitalWrite(YELLOW_LED, a);}
//  void green (boolean a)                       {digitalWrite(GREEN_LED,  a);}

//  #define yellowOn()                           digitalWrite(YELLOW_LED, true)
//  #define yellowOff()                          digitalWrite(YELLOW_LED, false)

  void blinkLed(uint8_t led, uint16_t del)       {for (int8_t i=1; i >= 0; i--) {digitalWrite(led, i); delay(del);}}
#else
  #define INIT_LEDS

  #define red(a)
  #define yellow(a)
  #define green(a)
  #define yellowOn()
  #define yellowOff()

  void blinkLed(uint8_t led, uint16_t del)       {delay(del);}
#endif // STATUS_LEDS

void flashYellow()                               {blinkLed(YELLOW_LED, 5);}
void blinkConfirm()                              {blinkLed(YELLOW_LED, 250);}

void flashRed()                                  {blinkLed(RED_LED, 10);}
void blinkWarning()                              {for (uint8_t i = 0; i < 6; i++) {blinkLed(RED_LED, 125);}}
void blinkWarningHalted()                        {lcdPrintProg1(haltMsg); while (true) {blinkLed(RED_LED, 125);}}

// ----------------

void pinOut(uint8_t pin)                         {pinMode(pin, OUTPUT); }
void pinIn (uint8_t pin)                         {pinMode(pin, INPUT); digitalWrite(pin, HIGH);} // set pinMode and enable internal pullups

void delay50()                                   {delay(50);}
void delayT()                                    {delay(delayTime * 10);}

uint8_t getConfig()                              {return (config >= NUM_CONFIGS)? 0 : config;}

void initVars(uint8_t len, int16_t arr[][3])     {for (int8_t i = 0; i < len; i++) {arr[i][0] = arr[i][1] = arr[i][2] = 1500;}}
void initVars(uint8_t len, int16_t arr[])        {for (int8_t i = 0; i < len; i++) {arr[i] = 1500;}}

// ----------------

void setup() {
#ifdef LOGGING  
  // "make sure that the default chip select pin is set to output, even if you don't use it:"
  pinOut(10); 
#endif

  INIT_LCD
  INIT_BACKLIGHT
  INIT_JOYSTICK
  INIT_LEDS
  INIT_APC220
  INIT_TEST
  
  initVars(4, servo       );
  initVars(8, motor       );
  initVars(8, rcHysteresis);

#ifdef LCD_ETPP
    lcdPrintProg0(intro);        // Print it only once.  The loop below will cause an ETPP to flicker and appear very dim
    delay(1600);
#else
  for (int8_t n = 0; n < 100; n++) {
    lcdPrintProg0(intro);
    delay(16);
    lcdClear();
  }
#endif

  restoreParams();

  serialBegin();

#ifdef LOGGING
  // see if the card is present and can be initialized:
  //digitalWrite(10, HIGH);     // maybe redundant
  boolean succ = SD.begin(CHIP_SELECT);
  setBacklight();
  if (!succ) {
    logMode = LOG_ERROR;
    resetLogMode();
    showLineBlink(noCard);
  }
#endif

  state = 0;
  initialized = false;
  lcdPrintProg1(reading);
  uint8_t readBytes = serialRead();
  initialized = true;

  if (logMode <= NO_LOGGING) {
    refreshDisplay();
    blinkConfirm();
  }
#ifdef LOGGING
  else {
    openLog();
  }
#endif
}

/* ------ */

void loop() {
  
#ifdef LOGGING
  if (logMode > NO_LOGGING) {
    doLog();
    refreshDisplay();
  }
#endif  

  int8_t statusFlag = pgm_read_byte_near(nextState[state] + 5); 
  if (statusFlag != -1) {
    // update the current values from MultiWii board and refresh display
    uint16_t current = millis();
    if (current - lastReadTime > delayTime * 10) {
      lastReadTime = current;
      serialRead();
      showStatusLine();
    }
  }
  
  int8_t key = getKeyPressed();
  if (key >=0) {
    int8_t next = getNextState(state, key);
    if (next >= 80) { // handle pseudo-states
      if (next <= 87) {
        enterPid(next-80); 
      }
      else if (next <= 92) {
        enterRates(next-90); 
      }
      else if (next == 93) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            enterValue(-state0-1);
            lcdPrintProg0(setMsgs[state0]);
          }
          key = getKeyPressed();
          if      (key == KEY_LEFT) { state0 = -1; }
          else if (key == KEY_UP  ) { state0 = (state0 + NUM_SETTINGS - 1) % NUM_SETTINGS; }
          else if (key == KEY_DOWN) { state0 = (state0                + 1) % NUM_SETTINGS; }
          else if (key >= 0)        { enterValue(state0+1); setBacklight(); }
        }
      } 
      else if (next == 94) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            lcdPrintProg0(profStates[state0]);
            lcdPrint(" Profile");
          }
          key = getKeyPressed();
          if      (key == KEY_LEFT) { state0 = -1; }
          else if (key == KEY_UP  ) { state0 = (state0 + 3) % 4; }
          else if (key == KEY_DOWN) { state0 = (state0 + 1) % 4; }
          else if (key >= 0) {
            if (state0 == 3) {
              extractProfile(defaultProfile);
            }
            else {
              lcdWrite(':'); 
              while (true) {
                if (key >= 0) {
                  profileMsg[2] = ('0' + profile/10);
                  profileMsg[3] = ('0' + profile%10);
                  getProfileName(profileMsg+6);
                  
                  lcdPrint(16, profileMsg);
                }
                key = getKeyPressed();
                if      (key == KEY_LEFT) { break; }
                else if (key == KEY_UP  ) { profile = (profile - 1 + NUM_PROFILES) % NUM_PROFILES; }
                else if (key == KEY_DOWN) { profile = (profile + 1)                % NUM_PROFILES; }
                else if (key >= 0) {
                  if      (state0 == 0) { restoreProfile(); } // load profile
                  else if (state0 == 1) { storeProfile();   } // save profile
                  else                  { deleteProfile();  } // clear profile
                  break;
                }
              }
            }
          }
        }
      } 
      else if (next == 96) {
        enterAuxState();
      }
      else if (next == 97) {
        int8_t offs = (key == KEY_UP)? 4 : 1;
        auxState = (auxState + offs) % 6;
      }
#ifdef LOGGING
      else if (next == 98) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            lcdPrintProg0(logStates[state0]);
            lcdPrint(" Log");
          }
          key = getKeyPressed();
          if      (key == KEY_UP)   {state0 = (state0 + 2) % 3;}
          else if (key == KEY_DOWN) {state0 = (state0 + 1) % 3;}
          else if (key == KEY_LEFT) {state0 = -1;}
          else if (key >= 0) {
            if (state0 == 0) {
              // start logging
              if (logMode >= NO_LOGGING && !offline) {
  #ifdef LOGVIEW
                logMode = confirm(formatMsg)? LF_LOGVIEW : LF_BINARY;
  #else
                logMode = LF_BINARY;
  #endif
                if (confirm(haltReqMsg)) {
                  storeParams();
                  setError(logEnbMsg);
                }
                openLog();
                state0 = -1;
              }
              else {
                blinkWarning();
              }
            }
            else if (state0 == 1) {
              playbackLog(); 
            }
            else {
              // delete log file
              int fileNo = 999;
              while (fileNo >= 0) {
                fileNo = selectLogfile(fileNo, false);
                if (fileNo >= 0) {
                  boolean succ = SD.remove(fileName);
                  setBacklight();
                  
                  if (succ) {
                    blinkConfirm();
                  }
                  else {
                    lcdPrintProg1(deleteFailed);
                    blinkWarning();
                    lcdPrintProg1(emptyLine);
                  }
                }
              }
            }
          }
        }
      }
#endif
      else if (next == 99) { 
#ifdef TEST
        if (key == KEY_SELECT) {
          config++;
          if (config >= NUM_CONFIGS) {
            config = 1;
          }
          lcdPrintProg1(configIds + configOffsets[getConfig()]);
          blinkConfirm();
        }
#endif          
        state = getConfig();         // select state depending on config
        if (state > 0) { 
          state += CONF_OFFSET; 
        } 
        else {
          state = 6; // unknown (offline)
        }
      }
      else if (next == 100) {
#ifdef CHECK_INPUT
        if (keyPressedTime > 10) {
          lcdClear();
          lcdPrintProg0(jsStatusMsg);

          while (true) {
            int8_t key = getKey();
            
            lcdWrite    (     16, keySymbols[key+1]);
  #ifdef KEY_PIN            
            format      ( 2 + 16, rawKey);
  #endif            
  #ifdef JOYSTICK            
            formatSigned( 7 + 16, jsHorAct  - jsHorCtr);
            formatSigned(12 + 16, jsVertAct - jsVertCtr);
  #endif
            key++;
            red   (key & 0x1);
            yellow(key & 0x2);
            green (key & 0x4);

            delay50();
          }
        }
        else
#endif
        {
          calibrate(calibMsgAcc, 'S');
          calibrate(calibMsgMag, 'E');
        }
      } 
      else if (next == 101) {
#ifdef MINMAX
  #ifdef GRAPHIC
        if (key == KEY_SELECT || keyPressedTime > 10) {
          graphMode = !graphMode;
        }
        else
  #endif
        {
          minMaxMode = (minMaxMode + 1) % 3; 
          showMinMax();
        }
#else        
  #ifdef GRAPHIC
        graphMode = !graphMode;
  #endif
#endif
      }
      next = state;
    }
    if (next >= 0 && logMode <= NO_LOGGING) {
      state = next;
      refreshDisplay();
    }
  }
}

void calibrate(prog_char const msg[], char ch) {
  if (confirm(msg)) {
    serialWrite(ch);
    state = 4;
    blinkLed(RED_LED, 200);
  }
}

void refreshDisplay() {
  lcdClear();

  // two levels of indirection...
  int8_t           index = pgm_read_byte_near(nextState[state] + 4);            // this is the index in the msgs array
  prog_char const* cptr0 = (prog_char const*) pgm_read_word_near(msgs + index); // this is the adress of the msg
                   index = pgm_read_byte_near(nextState[state] + 5); 
  prog_char const* cptr1 = (index >= 0)? (prog_char const*) pgm_read_word_near(msgs + index) : emptyLine;
  lcdPrintProg2(cptr0, cptr1);

  showStatusLine();
}

char getStatusChar(int8_t pres, int8_t mode) {return (pres)? (mode? '1' : '0') : '-';}

void showStatusLine() {
  if (state == 0) {
    lcdPrintProg1(configIds + configOffsets[getConfig()]);
    
    if (offline) {
      lcdPrintProg(9 + 16, offlineMsg);
    }
    else {
      lcdSetCursor(11 + 16);
      lcdWrite('v');      
      format(12 + 16, vers, 3, 2);
    }
  }
  else if (state == 1) {
    if (keyPressedTime > 10) {
      lcdPrintProg1(metaData);
      lcdWrite(baudRate + '0');
      enterKey();
#ifdef CREDITS
      lcdPrintProg1(credits);
      enterKey();
#endif      
    }
    char c0, c2;
    if (accPres > 1) {
      c0 = 'A';  // Acc
      c2 = 'c';
    }
    else {
      c0 = 'N'; // Nck - Nunchuk 
      c2 = 'k';
    }

    statusMsg[ 0] = c0;
    statusMsg[ 2] = c2;
    statusMsg[ 3] = getStatusChar(accPres , accMode );
    statusMsg[ 9] = getStatusChar(baroPres, baroMode);
    statusMsg[14] = getStatusChar(magPres , magMode );
    lcdPrint(16, statusMsg);
  }
  else if ((state >= 9 && state <= 13) || (state >= 37 && state <= 39)) {
    int8_t idx = (state == 9)? 0 : (state <= 13)? state - 10 : state - 33;

    format( 1 + 16, P8[idx], 4, 2);
    if (state != 39) {
      format( 6 + 16, I8[idx], 5, 4);
      if (state != 38) {
        formatSigned(12 + 16, -D8[idx], 3, 0);
      }
    }
  }
  else if (state == 14) { 
    // Dual Rate, Expo
    format( 0 + 16, rcRate8 * 2);
    format( 6 + 16, rcExpo8);
  }
  else if (state == 15) { 
    // Roll/Pitch Rate, Yaw Rate, Throttle PID
    format( 0 + 16, rollPitchRate);
    format( 6 + 16, yawRate);
    format(12 + 16, dynThrPID);
  }
  else if (state >= 16 && state <= 18) {  // Gyro, Acc, Mag
    lcdPrintProg(3, rollPitchYaw);
    for (int8_t i = 0; i < 3; i++) {
      int16_t val = (state == 16)? gyroData[i] MM_VAL : (state == 17)? accSmooth[i] MM_VAL : magADC[i] MM_VAL;
      formatSigned(i*6 + 16, val); 
    }
  }
  else if (state == 19) { // Angles
#ifdef GRAPHIC
    if (graphMode) {
      drawAngle(0);
      drawAngle(1);
    }
    else
#endif    
    {
      formatSigned(  1 + 16, angle[0] MM_VAL);
      formatSigned( 10 + 16, angle[1] MM_VAL);
    }
  }
  else if (state == 20) { // Alt, Head, Cycletime
    formatSigned( 0 + 16, altitude MM_VAL, 6, 0);
//    formatSigned( 0 + 16, altitude MM_VAL);
    formatSigned( 7 + 16, heading  MM_VAL);
    format      (12 + 16, cycleTime);
  }
  else if (state >= 21 && state <= 23) { // receiver input
    int8_t iOff = (state-21) * 3;
    int8_t sOff = 16;
    int8_t maxV = 3;

    if (state == 23) {
      sOff = 22;
      maxV = 2;
    }
    for (int8_t i = 0; i < maxV; i++) {formatScaled(i*6 + sOff, rcHysteresis[i + iOff] MM_VAL, !(i == 0 && state == 22));}
  }
  else if (state >= 24 && state < 35) {
    printEscLayout(esc_layouts[state-24]);
  }
  else if (state == 35) { // Battery Level
    format(     16, vbat,          4, 2);
    format( 5 + 16, powerMeterSum, 5, 0);
    format(11 + 16, powerTrigger,  5, 0);
    
  #ifdef STATUS_LEDS
    if (vbat > 0) {
      green (vbat >  VBATLEVEL1);      
      yellow(vbat >  VBATLEVEL3 && vbat <= VBATLEVEL0);
      red   (vbat <= VBATLEVEL2);
    }
  #endif
  }
  else if (state == 36) { // Activation Mode
    lcdPrintProg1(auxStates[auxState]);
    printAuxState(actMode[auxState]);
  }
}

void printEscLayout(uint8_t layout[]) {
  uint8_t tag = layout[0];
  
  if (tag == 0) {
    printEsc(     7, 20);
    printEsc(16 + 7, 21);
  }
  else if (tag == 1) {
    for (uint8_t i = 0; i < 4; i++) { printEsc(esc_pos_4[i], layout[i+1]); }
  }
  else if (tag == 2) {
    for (uint8_t i = 0; i < 6; i++) { printEsc(esc_pos_6[i], layout[i+1]); }
  }
  else {
    for (uint8_t i = 0; i < 8; i++) { printEsc(i*4, i); }
  }
}

void printEsc(uint8_t pos, uint8_t code) {
  if (code >= 20) {
    formatScaledSym(pos, servo[code-20] MM_VAL);  // servo signals - symmetrical
  }
  else {
    formatScaledAbs(pos, motor[code] MM_VAL);
  }
}

uint8_t auxPos[] = { 23, 24, 25, 29, 30, 31 };

void printAuxState(int8_t mode) {
  for (uint8_t i = 0; i < 6; i++) {
    char ch = (mode & auxMasks[i])? '*' : DOT;
    lcdWrite(auxPos[i], ch);
  }
}

int8_t getNextState(int8_t state, int8_t key) {
  int8_t col = (key == KEY_SELECT)? KEY_RIGHT : (key == STATUS)? STATUS-1 : key;

#ifdef MINMAX
  if (state >= 4 && state <= 6 && col == KEY_RIGHT) {
    showMinMax();
  }
#endif  

#ifdef STATUS_LEDS
  if (state == 35) {
    green(false); yellow(false); red(false);
  }
#endif

  int8_t next = pgm_read_byte_near(nextState[state] + col);

#ifdef LOGGING
  // disable "Logging" state when initialisation of SD Card failed 
  if (next == LOG_STATE && logMode == LOG_ERROR) {
    next = (key == KEY_UP)? 8 : 0;
  }
#endif

  return next;
}


/* ----- LCD Abstraction Layer: ----- */

#ifdef LCD_ETPP
// Eagle Tree Power Panel  http://www.eagletreesystems.com/MicroPower/9.htm
// The Power Panel is a physically small & light weight, 2x16, transflective (readable in sunlight) LCD that interfaces via i2c bus (two wire).
//   White wire  to Ground
//   Red wire    to +5V
//   Yellow wire to A4
//   Brown wire  to A5
// ETPP code contributed by Danal Estes

// *********************
// I2C general functions
// *********************
  #define I2C_SPEED 100000L     // 100kHz normal mode, this value must be used for a genuine WMP
  #define INTERNAL_I2C_PULLUPS  // Comment this if your hardware supplies i2c pullups outside the Amtel chip
  #define I2C_PULLUPS_ENABLE         PORTC |= 1<<4; PORTC |= 1<<5;   // PIN A4&A5 (SDA&SCL)
  #define I2C_PULLUPS_DISABLE        PORTC &= ~(1<<4); PORTC &= ~(1<<5);
  
  // Mask prescaler bits : only 5 bits of TWSR defines the status of each I2C request
  #define TW_STATUS_MASK	(1<<TWS7) | (1<<TWS6) | (1<<TWS5) | (1<<TWS4) | (1<<TWS3)
  #define TW_STATUS             (TWSR & TW_STATUS_MASK)
  
  void i2c_init(void) {
    #if defined(INTERNAL_I2C_PULLUPS)
      I2C_PULLUPS_ENABLE
    #else
      I2C_PULLUPS_DISABLE
    #endif
    TWSR = 0;        // no prescaler => prescaler = 1
    TWBR = ((16000000L / I2C_SPEED) - 16) / 2; // change the I2C clock rate
    TWCR = 1<<TWEN;  // enable twi module, no interrupt
  }
  
  void i2c_rep_start(uint8_t address) {
    TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN)  ;//| (1<<TWSTO); // send REAPEAT START condition
    waitTransmissionI2C(); // wait until transmission completed
    checkStatusI2C(); // check value of TWI Status Register
    TWDR = address; // send device address
    TWCR = (1<<TWINT) | (1<<TWEN);
    waitTransmissionI2C(); // wail until transmission completed
    checkStatusI2C(); // check value of TWI Status Register
  }
  
  void i2c_write(uint8_t data ) {	
    TWDR = data; // send data to the previously addressed device
    TWCR = (1<<TWINT) | (1<<TWEN);
    waitTransmissionI2C(); // wait until transmission completed
    checkStatusI2C(); // check value of TWI Status Register
  }
  
  uint8_t i2c_readAck() {
    TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);
    waitTransmissionI2C();
    return TWDR;
  }
  
  uint8_t i2c_readNak(void) {
    TWCR = (1<<TWINT) | (1<<TWEN);
    waitTransmissionI2C();
    return TWDR;
  }
  
  void waitTransmissionI2C() {
    uint8_t count = 255;
    while (count-->0 && !(TWCR & (1<<TWINT)) );
    if (count<2) { //we are in a blocking state => we don't insist
      TWCR = 0;  //and we force a reset on TWINT register
    }
  }
  
  void checkStatusI2C() {
    uint8_t tstat = TW_STATUS;
    if ( tstat == 0xF8) { //TW_NO_INFO : this I2C error status indicates a wrong I2C communication.
      exit (1);  //Fatal. 
    } 
  }

// *********************
// I2C Eagle Tree Power Panel primitives
// *********************
  void i2c_ETPP_init () {
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x00);            // ETPP command register
    i2c_write(0x24);            // Function Set 001D0MSL D : data length for parallel interface only; M: 0 = 1x32 , 1 = 2x16; S: 0 = 1:18 multiplex drive mode, 1×32 or 2×16 character display, 1 = 1:9 multiplex drive mode, 1×16 character display; H: 0 = basic instruction set plus standard instruction set, 1 = basic instruction set plus extended instruction set
    i2c_write(0x0C);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
    i2c_write(0x06);            // Cursor Move  000001IS I : 0 = DDRAM or CGRAM address decrements by 1, cursor moves to the left, 1 = DDRAM or CGRAM address increments by 1, cursor moves to the right; S : 0 = display does not shift,  1 = display does shifts
    lcdClear();         
  }

  void i2c_ETPP_send_cmd (byte c) {
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x00);            // ETPP command register
    i2c_write(c);
  }

  void i2c_ETPP_send_char (char c) {
    if (c > 0x0f) c |=  0x80;   // ETPP uses character set "R", which has A->z mapped same as ascii + high bit; don't mess with custom chars. 
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x40);            // ETPP data register
    i2c_write(c);
  }

  void i2c_ETPP_send_string (const char *s) {
    i2c_rep_start(0x76+0);      // I2C write direction
    i2c_write(0x40);            // ETPP data register
    while (*s) {byte c = *s++; if (c > 0x0f) c |=  0x80; i2c_write(c);}
  }

  void i2c_ETPP_set_cursor (byte col, byte row) {
    byte addr = col + row * 0x40;
    i2c_ETPP_send_cmd(0x80 | addr);     // High bit is "Set DDRAM" command, remaing bits are addr.  0x40 = first char of second line when in 2-line display (aka MUX 1:9 mode)
  }

  void i2c_ETPP_create_char (byte idx, uint8_t* array) {
    i2c_ETPP_send_cmd(0x80);             // CGRAM and DDRAM share an address register, but you can't set certain bits with the CGRAM address command.   Use DDRAM address command to be sure high order address bits are zero. 
    i2c_ETPP_send_cmd(0x40 | byte(idx * 8));   // Set CGRAM address 
    i2c_rep_start(0x76+0);               // I2C write direction
    i2c_write(0x40);                     // ETPP data register
    for (byte i = 0 ; i<8 ; i++) {i2c_write(*array); array++;}
  }

// *********************
// LCD functions similar to LiquidCrystal lib functions, as adapted for ETPP
// *********************
  void lcdClear() {
    i2c_ETPP_send_cmd(0x01);                              //Clear display command, which does NOT clear an Eagle Tree because character set "R" has a '>' at 0x20
    for (byte i = 0; i<80; i++) i2c_ETPP_send_char(' ');  // Blanks for all 80 bytes of RAM in the controller, not just the 2x16 display
  }
  
  void lcdBlink()                                                      {
    i2c_ETPP_send_cmd(0x0D);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
    yellowOn();
  }
  void lcdNoBlink()                                                      {
    i2c_ETPP_send_cmd(0x0C);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
  }
  void lcdCursor()                                                      {
    i2c_ETPP_send_cmd(0x0E);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
    yellowOn();
  }
  void lcdNoCursor()                                                      {
    i2c_ETPP_send_cmd(0x0C);            // Display on   00001DCB D : 0 = Display Off, 1 = Display On; C : 0 = Underline Cursor Off, 1 = Underline Cursor On; B : 0 = Blinking Cursor Off, 1 = Blinking Cursor On
  }
  void lcdCreateChar(int8_t idx, uint8_t* array)                       {i2c_ETPP_create_char(idx, array);}
  void lcdSetCursor(uint8_t pos)                                       {i2c_ETPP_set_cursor(pos % 16, pos / 16);}
  void lcdWrite(int8_t digit)                                          {i2c_ETPP_send_char(digit);}
  void lcdPrint(char const      msg[])                                 {i2c_ETPP_send_string(msg);}

#elif defined(SERIAL_LCD)

  void lcdCmd(byte arg)                                                {lcd.print(0xFE, BYTE); lcd.print(arg, BYTE); lcdDelay();}
  void lcdClear()                                                      {lcdCmd(0x01);}
  void lcdBlink()                                                      {lcdCmd(0x0D); yellowOn();}
  void lcdNoBlink()                                                    {lcdCmd(0x0C);}
  void lcdCursor()                                                     {lcdCmd(0x0E); yellowOn();}
  void lcdNoCursor()                                                   {lcdCmd(0x0C);}
  void lcdCreateChar(int8_t idx, uint8_t* array)                       {}
  void lcdSetCursor(uint8_t pos)                                       {if (pos>15) pos += 48; lcdCmd(128+pos);}
  void lcdWrite(int8_t digit)                                          {lcd.print(digit, BYTE);}
  void lcdPrint(char const msg[])                                      {lcd.print(msg);}

#else  // next set of functions work for LiquidCrystal and SRLCD libraries

  void lcdClear()                                                      {lcd.clear();}
  void lcdBlink()                                                      {lcd.blink();  yellowOn();}
  void lcdNoBlink()                                                    {lcd.noBlink();}
  void lcdCursor()                                                     {lcd.cursor(); yellowOn();}
  void lcdNoCursor()                                                   {lcd.noCursor();}
  void lcdCreateChar(int8_t idx, uint8_t* array)                       {lcd.createChar(idx, array);}
  void lcdSetCursor(uint8_t pos)                                       {lcd.setCursor(pos % 16, pos / 16);}
  void lcdWrite(int8_t digit)                                          {lcd.write(digit);}
  void lcdPrint(char const msg[])                                      {lcd.print(msg);}
#endif

void lcdPrint     (uint8_t pos, char const      msg[])                 {lcdSetCursor(pos); lcdPrint(msg);}
void lcdPrintProg (uint8_t pos, prog_char const msg[])                 {lcdSetCursor(pos); lcdWriteProg(msg);}

void lcdPrintProg0(prog_char const line[])                             {lcdPrintProg( 0, line);}
void lcdPrintProg1(prog_char const line[])                             {lcdPrintProg(16, line);}
void lcdPrintProg2(prog_char const line0[], prog_char const line1[])   {lcdPrintProg0(line0); lcdPrintProg1(line1);}

void lcdWrite(uint8_t pos, int8_t digit)                               {lcdSetCursor(pos); lcdWrite(digit);}

void lcdWriteProg(prog_char const msg[]) {
  char* lptr = lineBuffer;
  char  last = -1;
  for (int8_t i = 0, k = 0; last != '\0' && k < 16; i++) {
    last = pgm_read_byte_near(msg + i);
    if (last <= 16) {
      // a char ch < 16 represent empty strings of length ch - this saves a lot of memory
      for (int8_t j = 0; j < last && k < 16; j++) {*lptr++ = ' '; k++;}
    }
    else {
      *lptr++ = last;
      k++;
    }
  }
  *lptr++ = '\0';

  lcdPrint(lineBuffer);
}

// ----------------

void enterPid(int8_t index) {
  static uint8_t col [] = { 5, 11, 15, };
  static uint8_t maxv[] = { 250, 250, 40, };
  
  int8_t idx = index - 1;
  if (idx < 0) {idx = 0;}
 
  uint8_t val[3];
  val[0] = P8[idx]; 
  val[1] = I8[idx]; 
  val[2] = D8[idx];
  
  int8_t local    =  0; // local state (-1..4)
  int8_t lastKey  = -1;
  int    count    =  0;
  int    repeat   =  0;
  int8_t maxState = (index < 6)? 3 : 8 - index;
  
  lcdBlink();
  while (local >= 0 && local < maxState) {
    lcdSetCursor(col[local] + 16);

    int8_t key = getKey();
    if (key < 0) {
      lastKey = key;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_LEFT) {
        local--;
      }
      else if (key == KEY_SELECT || (key == KEY_RIGHT && local == maxState-1)) {
        uint8_t pp = val[0];
        uint8_t ii = val[1];
        uint8_t dd = val[2];

        if (index == 0) {
          P8[1] = pp;
          I8[1] = ii;
          D8[1] = dd;
        }
        P8[idx] = pp;
        I8[idx] = ii;
        D8[idx] = dd;
        
        serialWrite();
        local = -1;
      }
      else if (key == KEY_RIGHT) {
        local++; 
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          int tmp = val[local];
          if (key == KEY_UP) {
            if (tmp > 0) {
              tmp--;
            }
          }
          else {
            if (tmp < maxv[local]) {
              tmp++;
            }
          }
          val[local] = tmp;
          if      (local == 0) { format      ( 1 + 16,  tmp, 4, 2); }
          else if (local == 1) { format      ( 6 + 16,  tmp, 5, 4); }    
          else if (local == 2) { formatSigned(12 + 16, -tmp, 3, 0); } 
          repeat++;
        }
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
}

void enterRates(int8_t index) {
  int8_t minState = 0;
  int8_t maxState = 3;

  uint16_t v0, v1, v2;
  if (index == 0) {
    v0 = rcRate8;
    v1 = rcExpo8;
    maxState = 2;
  }
  else if (index == 1) {
    v0 = rollPitchRate;
    v1 = yawRate;
    v2 = dynThrPID;
  }
  else {
    v2 = powerTrigger;
    minState = 2;
  }

  uint8_t val[3];
  val[0] = v0;
  val[1] = v1;
  val[2] = v2;
  
  int8_t local   =  minState;  // local state (-1..4)
  int8_t lastKey = -1;
  int    count   =  0;
  int    repeat  =  0;

  lcdBlink();
  while (local >= minState && local < maxState) {
    uint8_t col = local * 6 + 4 + 16;
    if (col > 31) {col = 31;}
    lcdSetCursor(col);
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_LEFT) {
        local--;
      }
      else if (key == KEY_SELECT || (key == KEY_RIGHT && local == maxState-1)) {
        v0 = val[0];
        v1 = val[1];
        v2 = val[2];
        
        if (index == 0) {
          rcRate8       = v0;
          rcExpo8       = v1;
        }
        else if (index == 1) {
          rollPitchRate = v0;
          yawRate       = v1;
          dynThrPID     = v2;
        }
        else {
          powerTrigger  = v2;
        }

        serialWrite();
        local = -1;
      }
      else if (key == KEY_RIGHT) {
        local++; 
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          uint8_t tmp = val[local];
          if (index < 2) {
            if (key == KEY_UP) {
              if (tmp > 0) {
                tmp--;
              }
            }
            else {
              uint8_t maxVal = (index == 0 && local != 1)? 250 : 100;
              if (tmp < maxVal) {
                tmp++;
              }
            }
          }
          else {
            if (key == KEY_UP) {
              tmp -= 16;
            }
            else {
              tmp += 16;
            }
          }
          val[local] = tmp;
          if (index == 0 && local == 0) {
            tmp *= 2;
          }
          format(local * 6 + 16, tmp);
          repeat++;
        }
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
}

boolean enterName(char name[]) {
  int8_t local   =  0; // local state (-1..8)
  int8_t lastKey = -1;
  int    count   =  0;
  int    repeat  =  0;
  int8_t maxCh   =  sizeof(charsInName)-2;
  
  for (int8_t i = 0; i < 8; i++) {
    if (name[i] == '-') {
      name[i] = ' ';
    }
  }
  
  int8_t idx = findChar(name[0]);
  
  lcdCursor();
  lcdWrite    (6+16, name[0]);
  lcdSetCursor(6+16);
  
  while (local >= 0 && local < 8) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      boolean refresh = false;
      if (key == KEY_LEFT) {
        local--;
        refresh = (local >= 0);
        if (refresh) {
          idx = findChar(name[local]);
        }
      }
      else if (key == KEY_RIGHT) { 
        local++; 
        refresh = (local < 8);
        if (refresh) { // handle RIGHT in third column as SELECT
          idx = findChar(name[local]);
        }
      }
      else if (key == KEY_SELECT) {
        local = 8;
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          if (key == KEY_DOWN && idx < maxCh) {
            idx++;
            refresh = true;
          }
          else if (key == KEY_UP && idx > 0) {
            idx--;
            refresh = true;
          }
          if (refresh) {
            name[local] = charsInName[idx];
          }
        }
      }

      if (refresh) {
        lcdWrite    (local + 6 + 16, name[local]);
        lcdSetCursor(local + 6 + 16);
        repeat++;
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
  
  return (local == 8);
}

int8_t findChar(char ch) {
  int8_t res = 0;
  int8_t len = sizeof(charsInName)-1;
  for (int8_t i = 1; res < 1 && i < len; i++) {
    if (charsInName[i] == ch) {
      res = i;
    }
  }
  return res;
}

/* ----- */

void enterAuxState() {
  lcdCursor();
  
  int8_t local = 0; // local state (-1..6)
  int8_t mode  = actMode[auxState];
  int8_t key   = 0;

  while (local >= 0 && local < 6) {
    if (key >= 0) {
      printAuxState(mode);
      lcdSetCursor(auxPos[local]);
    }
    key = getKeyPressed();
    if      (key == KEY_LEFT)   {local--;}
    else if (key == KEY_RIGHT)  {local++;}
    else if (key == KEY_SELECT) {local = 6;}
    else if (key >= 0)          {mode ^= auxMasks[local];}

    delay50();
  }
  waitForKeyUp();
  
  if (local == 6) {
    actMode[auxState] = mode;
    serialWrite();    
  }
}

/* ----- */

// !!! remove Version - index 0    

void enterValue(int8_t index) {  
  bool showValue = (index < 0);
  if (showValue) {
    index = -index;
  }
  else {
    lcdWrite(':');
    lcdBlink();
  }
  
  uint16_t val = backlight;
  if      (index == 1) { val = delayTime; lcdPrintProg1(delaySuff); }
  else if (index == 2) { val = baudRate; }

  int    repeat  =  0;
  int    count   =  0;
  int8_t lastKey = -1;

  while (true) {
    if      (index == 1) { format( 5 + 16, val * 10); }
    else if (index == 2) { format( 5 + 16, baudRates[val]*12); lcdPrint("00"); }
    else                 { format( 5 + 16, val); }
    
    if (showValue) {
      return;
    }
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  =  0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_UP || key == KEY_DOWN) {
        if (repeat > 3 || (count % 8) == 0) {
          int8_t  incr = 1;
          uint8_t maxV = 250;

          if   (index == 2) { maxV = 8; }
          else              { incr = 5; }

          if (key == KEY_UP) { 
            val = (val < incr)? 0 : val - incr;
          }
          else {
           val = (val > maxV - incr)? maxV : val + incr; 
          }

          if (index == 3) {
            setBacklight(val);
          }
          repeat++;
        }
      }
      else {
        if (key != KEY_LEFT) {
          if      (index == 1) { delayTime    = val; }
          else if (index == 2) { baudRate     = val; }
          else                 { backlight    = val; }
  
          storeParams();
        }
        waitForKeyUp();
        return; 
      }
    }
    count++;
    delay50();
  }
}

// ----------------

void format      (uint8_t pos, int val)                          {format      (pos, val, 4,   0);}
void format      (uint8_t pos, int val, int8_t len, int8_t prec) {formatNum   (pos, val, len, prec, false);}

void formatSigned(uint8_t pos, int val)                          {formatSigned(pos, val, 4,   0);}
void formatSigned(uint8_t pos, int val, int8_t len, int8_t prec) {formatNum   (pos, val, len, prec, true);}

void formatNum   (uint8_t pos, int val, int8_t len, int8_t prec, bool sgn) {
  char temp[len+1];
  int8_t low = 0;
  if (sgn) {
    temp[0] = (val > 0)? '+' : (val < 0)? '-' : ' ';
    low++;
  }
  val = abs(val);
  for (int8_t i = len-1; i >= low; i--) {
    if (len-i == prec) {
      temp[i] = '.';
    }
    else {
      temp[i] = '0' + (val % 10);
      val    /= 10;
    }
  }
  if (val > 0) { // overflow
    for (int8_t i = len-1; i >= 0; i--) {
      temp[i] = '#';
    }
  }
  temp[len] = '\0';
  if (pos < 32) {
    lcdPrint(pos, temp);
  }
  else {
    lcdPrint(temp);
  }
}

void formatScaledSym(uint8_t pos, int val) {formatScaled(pos, val, true);}
void formatScaledAbs(uint8_t pos, int val) {formatScaled(pos, val, false);}

void formatScaled   (uint8_t pos, int val, boolean sym) { // sym: -100..+100, !sym=0..100
  if (sym) {
    val = (val - 1500)/5;
  }
  else {
    val = (val - 1000)/10;
  }  
#ifdef GRAPHIC
  if (graphMode) {
#ifdef GRAPHIC
    if (sym) {
      val = (val + 100)/2;
    }
#endif
    formatGraphic(pos, val);
  }
  else
#endif
  {
    formatSigned(pos, val);
  }
}

#ifdef GRAPHIC
static int8_t charsInitialized;    // chars for servo signals are initialized
  
static byte bars[8] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x15, };

void formatGraphic(uint8_t pos, int8_t val) { // 0..100
  if (charsInitialized != 1) {
    charsInitialized = 1;

    static byte less[8] = { 0x00, 0x04, 0x0C, 0x1C, 0x0C, 0x04, 0x00, 0x15, };
    static byte grt [8] = { 0x00, 0x04, 0x06, 0x07, 0x06, 0x04, 0x00, 0x15, };

    byte pattern = 0x10;
    for (int8_t i = 0; i <= 5; i++) {
      for (int8_t j = 0; j < 7; j++) {
        bars[j] = pattern;
      }
      lcdCreateChar(i, bars);
      pattern >>= 1;
    }
    lcdCreateChar(6, less);
    lcdCreateChar(7, grt);
  }  
  
  static char bar[4];
  for (int8_t i = 0; i < 4; i++)    { bar[i] = 5; }
  
  if      (val < -100 || val > 200) { bar[0] = 6; bar[3] = 7; }    // invalid
  else if (val <    0)              { bar[0] = 6; }                // <...
  else if (val >= 100)              { bar[3] = 7; }                // ...>
  else                              { bar[val/25] = (val%25)/5; }  // ..|.

  lcdSetCursor(pos);
  for (int8_t i = 0; i < 4; i++) {
    lcdWrite(bar[i]); 
  }
}

#if defined(TEXTSTAR_GRAPHIC)
void formatBar(uint8_t len, int8_t val) { // 0..100
  if (charsInitialized != 2) {
    charsInitialized = 2;

    bars[0] = 0x15;
    for (int8_t j = 1; j < 7; j++) {
      bars[j] = 0;
    }
    lcdCreateChar(0, bars);

    bars[3] = 0x10;
    bars[4] = 0x10;
    lcdCreateChar(1, bars);
    
    byte pattern = 0x10;
    for (int8_t i = 2; i <= 6; i++) {
      for (int8_t j = 2; j < 6; j++) {
        bars[j] = pattern;
      }
      lcdCreateChar(i, bars);
      pattern >>= 1;
      pattern |= 0x10;
    }
  }

  int act = (val * len * 6) / 100;
  for (int8_t i = 0, low = 0; i < len; i++, low += 6) {
    char ch = 0;
    if (act >= low + 6) {
      ch = 6;
    }
    else if (act >= low) {
      ch = act - low;
    }
    lcdWrite(ch);
  }
}
#endif

static uint8_t angleChars[4][8];

int8_t round8(float val) {return (int8_t) (val + 0.5f);}

void drawAngle(int8_t index) {

  // clear char arrays
  uint8_t *cptr = angleChars[0];
  for (int8_t i = 0; i < 32; i++) {
    *cptr++ = 0;
  }

  int8_t  charOffset = index * 4;
  int16_t alpha      = - angle[index] MM_VAL;
  if (alpha >= -360 && alpha <= 360) {
    float ang  = radians(alpha);
    float sinA = sin(ang);
    float cosA = cos(ang);
    
  //  if (abs(sinA) > 0.25f) {
  //    angleChars[0][4] = 0x14;  // mark horizon
  //    angleChars[3][4] = 0x05;
  //  }
    
    int8_t x0 = round8(- cosA * 10.0f);
    int8_t y0 = round8(- sinA * 10.0f);
    drawLine(x0, y0, -x0, -y0);
    
    int8_t x1 = round8(- sinA * 3.0f);
    int8_t y1 = round8(  cosA * 3.0f);
    drawLine(0, 0, x1, y1);
  }
  else {
    angleChars[1][4] = 0x06;  
    angleChars[2][4] = 0x18;  
  }
  
  for (int8_t i = 0; i < 4; i++) {
    lcdCreateChar(charOffset + i, angleChars[i]);
    lcdWrite(16 + 2 + index * 8 + i, charOffset + i);
  }
  
  charsInitialized = 0;  
}

void drawLine(int8_t x0, int8_t y0, int8_t x1, int8_t y1) {
  // Bresenham again...
  
  int8_t dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
  int8_t dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
  int8_t err = dx+dy, e2; /* error value e_xy */
 
  for(;;){  /* loop */
    setPixel(x0, y0);
    if (x0==x1 && y0==y1) break;
    e2 = 2*err;
    if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
    if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
  }  
}

void setPixel(int8_t x, int8_t y) { 
  // x: -10..+9, y: -3..+4
  
  if (x >= -10 && x < 10 && y >= -3 && y < 5) {
    int8_t nrow = 4 - y;   // 0..7
    int8_t ncol = x + 10;  // 0..19
    int8_t idx  = ncol / 5;
    int8_t pcol = ncol % 5;
    
    angleChars[idx][nrow] |= (0x10 >> pcol);
  }
}
#endif // GRAPHIC

// ----------------

#ifdef MINMAX
void showMinMax() {
  lcdClear();
  lcdPrintProg0(minMaxDsc + 4 * minMaxMode);
  blinkConfirm();
}
#endif // MINMAX

/* ----- */

#ifdef EMULATE_TEXTSTAR
void emulateLcd() {
  lcdPrintProg1(emuLcdMsg);
  delay(2000);
  lcdClear();
  
  uint8_t bsize;
  uint8_t state = 0;
  while (true) {
    if (serialAvailable()){
      uint8_t next = serialReadByte();
      if (state == 0) {
        if      (next == 0xfe) {state = 1;       }  // control character escape sequence
        else if (next == 0x0c) {lcdClear();      }
        else if (next == 0x0d) {lcdSetCursor(16);}
        else                   {lcdWrite(next);  }
      }
      else if (state == 1) {
        if      (next == '+') {state = 2;} // cursor BlinkMode ON
        else if (next == 'L') {state = 3;} // set cursor position
        else if (next == 'b') {state = 4;} // bar graphic
        else                  {state = 0;} // ignore
      }
      else if (state == 2) {
        lcdBlink();
        state = 0;
      }
      else if (state == 3) {
        if (next == 0x01) {lcdSetCursor( 0);}
        else              {lcdSetCursor(16);}
        state = 0;
      }
      else if (state == 4) {
        // next is size of output field
        bsize = next;
        state = 5;
      }
      else if (state == 5) {
        // next is supposed to be a value in percent (?)
#if defined(GRAPHIC) && defined(TEXTSTAR_GRAPHIC)
        formatBar(bsize, next);
#else       
        formatNum(100, (int8_t) next, bsize, 0, true);
#endif
        state = 0;
      }
    }
    else {
      // in order to allow switching of Telemetry displays, the refresh rate of
      // the telemetry should be set to 1 Hz in the MWC sketch (otherwise, it
      // ignores sent characters because the Serial is always busy).

      flashYellow();
      int8_t key = getKeyPressed();
      if (key >= 0) {
        char                          ch = 'B';   // LCD_VALUE_DOWN
        if      (key == KEY_LEFT)   { ch = 'A'; } // LCD_VALUE_PREV
        else if (key == KEY_RIGHT)  { ch = 'C'; } // LCD_VALUE_NEXT
        else if (key == KEY_UP)     { ch = 'D'; } // LCD_VALUE_UP
        else if (key == KEY_SELECT) { break; }

        serialWrite(ch);                          // key down
        waitForKeyUp();
        serialWrite(ch+32);                       // key up
        flashRed();
      }
      delay50();
    }
  }
  waitForKeyUp();
}
#endif


/* ----- Input Devices Abstraction Layer: ----- */

boolean confirm(prog_char const msg[]) {
  lcdPrintProg2(msg, yesNoMsg);
  waitForKeyUp();
  yellowOn();
    
  int8_t key = -1;
  while (key != KEY_LEFT && key != KEY_RIGHT) {
    key = getKeyPressed();
  }      
  waitForKeyUp();
  return (key == KEY_RIGHT);
}

int8_t enterKey() {
  yellowOn();
  int8_t key = -1;
  while ((key = getKeyPressed()) < 0);
  yellowOff();
  return key;
}

int8_t getKeyPressed() {
  uint8_t cnt = 0;
  int8_t  key = getKey();
  if (key >= 0) {
    while (rawGetKey() >= 0) {
      if (cnt < 255) {
        cnt++;
      }
      if (cnt > 20) {
        flashRed();
      }
      delay50();
    }
  }
  keyPressedTime = cnt;
  return key;
} 

void waitForKeyUp() {
  while (rawGetKey() >= 0);

  yellowOff();
  lcdNoCursor();
  lcdNoBlink();
}

int8_t getKey() {
  static int8_t oldKey = -1;

  int8_t key = rawGetKey();
  if (key != oldKey) {
    delay50();		// wait for debounce time
    key = rawGetKey();
    if (key != oldKey) {			
      oldKey = key;
    }
  }
  return key;
}

#ifndef HWKEYS
//static int adc_key_val[NUM_KEYS] = { 30, 175, 360, 535, 760 }; // thresholds of the cascading resistors
static int adc_key_val[NUM_KEYS] = { 64, 220, 390, 560, 860 }; // thresholds of the cascading resistors
#endif

int8_t rawGetKey() {
  
#ifdef HWKEYS
  if      (digitalRead(HW_RIGHT) ==0) {return KEY_RIGHT; }
  else if (digitalRead(HW_UP)    ==0) {return KEY_UP;    }
  else if (digitalRead(HW_DOWN)  ==0) {return KEY_DOWN;  }
  else if (digitalRead(HW_LEFT)  ==0) {return KEY_LEFT;  }
  else if (digitalRead(HW_SELECT)==0) {return KEY_SELECT;}
#else
  #if defined(KEY_PIN)
  rawKey = analogRead(KEY_PIN);   // read the value from the sensor  
  for (int8_t k = 0; k < NUM_KEYS; k++) {
    if (rawKey < adc_key_val[k]) {
      return k;
    }
  }
  #endif
#endif

#ifdef JOYSTICK
 #ifdef JS_SELECT_PIN
  int sel  = analogRead(JS_SELECT_PIN);
  if (sel < 100) {
    return KEY_SELECT;
  }
 #endif
 
  jsHorAct  = analogRead(JS_HOR_PIN);
  jsVertAct = analogRead(JS_VERT_PIN);

  if (jsHorAct < 1000 || jsVertAct < 1000) { // is joystick present? (int. pullups are on)
    int hor  = (jsHorAct  - jsHorCtr ) * JS_HOR_SIGN;
    int vert = (jsVertAct - jsVertCtr) * JS_VERT_SIGN;

    if (vert < JS_DEADBAND && vert > -JS_DEADBAND) {
      if      (hor < -5 * JS_DEADBAND)  {return KEY_RIGHT;}
      else if (hor >  5 * JS_DEADBAND)  {return KEY_LEFT; }
    }
    else if (hor < JS_DEADBAND && hor > -JS_DEADBAND) {
      if      (vert < -5 * JS_DEADBAND) {return KEY_UP;   }
      else if (vert >  5 * JS_DEADBAND) {return KEY_DOWN; }
    }
  }
#endif
  return -1;
}


/* ----- Serial Communication: ----- */

#define BUFFER_SIZE 128           // max. number of bytes read from serial

static uint8_t* bptr;
static uint8_t  buffer[ BUFFER_SIZE ];

boolean serialAvailable()         {return Serial.available();}
void    serialBegin()             {Serial.begin(long(baudRates[baudRate])*1200);}  
void    serialFlush()             {Serial.flush();}
uint8_t serialReadByte()          {Serial.read();}  
void    serialWrite(uint8_t ch)   {Serial.write(ch);}  
//void serialWriteBuffer()          {Serial.write(buffer, bptr - buffer); serialFlush();}

void serialize8   ( int8_t a)     {*bptr++ = a;}
void serialize16  (int16_t a)     {*bptr++ = a; *bptr++ = a/256;}

void deserialize8 (  int8_t& val) {val = *bptr++;}
void deserialize8 ( uint8_t& val) {val = *bptr++;}
void deserialize16( int16_t& val) {uint8_t low = *bptr++; uint8_t hi = *bptr++; val = (int16_t)  (hi * 256 + low);}
void deserialize16(uint16_t& val) {uint8_t low = *bptr++; uint8_t hi = *bptr++; val = (uint16_t) (hi * 256 + low);}

void deserialize16(int16_t arr[]) {
  deserialize16(arr[0]); 
  if (initialized) {
    arr[1] = min(arr[0], arr[1]); 
    arr[2] = max(arr[0], arr[2]);
  }
  else {
    arr[1] = arr[2] = arr[0];
  }
}

void setOffline(prog_char const* msg) {
  lcdPrintProg1(msg);
  blinkWarning();
  offline = true;
  vers    = 18;

  resetLogMode();
  refreshDisplay();
}

void setError(prog_char const msg[]) {
  lcdPrintProg0(msg);
  blinkWarningHalted();
}

void setFrameError(uint8_t code, int8_t arg) {
  resetLogMode();             // stop logging

#ifdef DUMP_BUFFER
  dumpBuffer();
#endif  

  lcdPrintProg0(wrongMsg);
  lcdPrintProg (6, frameError[code]);
  format(13, arg, 3, 0);
  
  blinkWarningHalted();
}

/* ----- */

#ifdef DUMP_BUFFER
void dumpBuffer() {
  // read buffer
  while (serialAvailable() && (bptr - buffer) < BUFFER_SIZE) {
    *bptr++ = serialReadByte();
  }
  
  // print read bytes
  uint8_t avail = (bptr - buffer);
  for (uint8_t i = 0; i < avail; i += 8) {
    lcdClear();
    format( 0, i  , 2, 0);
    format(16, i+4, 2, 0);
    for (uint8_t j = 0; j < 8; j++) {
      if (i + j < avail) {
        uint8_t pos = 4 + j * 3 + 4 * (j > 3);
        uint8_t tmp = buffer[i + j];
        uint8_t hi  = tmp / 16;
        uint8_t low = tmp % 16;
        char    ch  = '0' + hi  + 7 * (hi  > 9);
        char    cl  = '0' + low + 7 * (low > 9);
        lcdWrite(pos, ch);
        lcdWrite(cl);
      }
    }
    int8_t key;
    while ((key = getKeyPressed()) < 0);
    waitForKeyUp();

    if (key == KEY_UP) {
      i -= 16;
    }
  }
}
#endif

/* ----- */

#ifndef TEST
uint8_t serialRead() {
  uint16_t readBytes = 0;

  if (!offline) {
    serialFlush();

    char     tag   = 'M';
    int16_t  retry = 600; // 600 * 50ms = 30 sec timeout
    uint16_t len   = 3;
    uint8_t  led   = RED_LED;
    
    bptr = buffer;
    while (readBytes < len && retry > 0) {
      if (!initialized) {
        format(9 + 16, readBytes, 3, 0);
        
        int8_t key = getKey();
        switch (key) {
#ifdef EMULATE_TEXTSTAR
          case KEY_RIGHT:
            emulateLcd();  // no break!
#endif            
          case KEY_LEFT:
            // cancel loop
            setOffline(cancelRead);
            return 0;
        }
      }
      if (serialAvailable()) {
        uint8_t last = serialReadByte();
        *bptr++      = last;
        led          = YELLOW_LED;
        readBytes++;
        
        if (readBytes == 1) {
          if (last != tag) {
            setFrameError(1, last);  // wrong prefix
          }
        }
        else if (readBytes == 2) {
          last -= 18; // v1.8
          if (last > sizeof(frameLength)) {
            setFrameError(4, last); // wrong version
          }
          len = frameLength[last];
        }
        else if (readBytes == len) {
          if (last != tag) {
            setFrameError(2, last); // wrong suffix
          }
        }
      }          
      else {
        int8_t mod = retry % 20;
        if (readBytes < 1 && mod == 0) {
          serialWrite(tag);
        }
#ifdef STATUS_LEDS        
        if (mod == 1) {
          blinkLed(led, 10);    
        }
#endif
        delay50();
        retry--;
      }
    }
    
    if (readBytes == 0) {
      setOffline(timeoutError);
    }
    else if (readBytes < len) {
      setFrameError(0, readBytes);
    }
    else {
      deserialize();
      if (config >= NUM_CONFIGS) {
        setFrameError(3, config);
      }
    }
  }
  else {
    flashRed();
  }
  return readBytes;
}
#else // TEST
uint8_t serialRead() {
  bptr = buffer;
  int8_t i;

  serialize8('M');                                      // 00: tag
  serialize8(18);                                       // 01: version

  for (i = 0; i < 3; i++) serRnd(accSmooth[i]);         // 02-07: acc
  for (i = 0; i < 3; i++) serRnd(gyroData [i]);         // 08-13: gyro
  for (i = 0; i < 3; i++) serRnd(magADC   [i]);         // 14-19: mag

  serialize16(33);                                      // 20-21: altitude
  serialize16(66);                                      // 22-23: heading

  for (i = 0; i < 4; i++) serRnd(servo       [i]);     // 24-31: servos
  for (i = 0; i < 8; i++) serRnd(motor       [i]);     // 32-47: motor
  for (i = 0; i < 8; i++) serRnd(rcHysteresis[i]);     // 48-63: receiver

  serialize8(1);                                       // 64: avail. sensors:   0-nunchukPres, 1-accPres, 2-baroPres, 3-magPres
  serialize8(1);                                       // 65: sensor modes:     0-accMode, 1-baroMode, 2-magMode

  serialize16(5120);                                   // 66-67: cycle time
  
  for (i = 0; i < 2; i++) {
    serRnd(angle[i]);                                  // 68-71: roll and pitch angle
  }
  
  serialize8(2);                                       // 72: config / multitype (Tri, Quad, ...)

  for (i = 0; i < 5; i++) {
    serialize8(40); serialize8(30); serialize8(13);    // 73-87: P-I-D (roll, pitch, yaw, alt, vel)
  }
  serialize8(140); serialize8(45);                     // 88-89: pLevel, iLevel
  serialize8(0);                                       // 90:    pMag
  
  serialize8(50);                                      // 91: rcRate (dual rate)
  serialize8(66);                                      // 92: rcExpo
  serialize8(25);                                      // 93: roll/pitch rate
  serialize8(50);                                      // 94: yaw rate
  serialize8(75);                                      // 95: dyn. throttle PID
  
  for (i = 0; i < 6; i++) {                            // 96-101: activation modes (level, baro, mag, camStab, camTrig, arming)
    serialize8(i);
  }

  serialize16(3000);                                   // 102-103: powerMeterSum
  serialize16(40000);                                  // 104-105: powerTrigger
  serialize8(90 + random(30));                         // 106: vbat (battery level)

  serialize16(10);                                     // 107-114: 4 16bit debug values
  serialize16(100);                                  
  serialize16(1000);                                   
  serialize16(10000);                                  
  
  serialize8('M');                                      // 115 - tag
    
  deserialize();
  
  if (state != 35) {
    flashYellow();
  }
  delay(200);

  return (bptr - buffer);
}

void serRnd(int16_t val MM_DIM) {serialize16(val MM_ACT + random(20) - 9);}
#endif

void deserialize() {
  bptr = buffer + 1;
  int8_t i;

  deserialize8(vers);                                         // 01: version

  for (i = 0; i < 3; i++) deserialize16(accSmooth[i]);        // 02-07: acc
  for (i = 0; i < 3; i++) deserialize16(gyroData [i]);        // 08-13: gyro
  for (i = 0; i < 3; i++) deserialize16(magADC   [i]);        // 14-19: mag

  deserialize16(altitude);                                    // 20-21: altitude
  deserialize16(heading);                                     // 22-23: heading

  for (i = 0; i < 4; i++) deserialize16(servo       [i]);     // 24-31: servos
  for (i = 0; i < 8; i++) deserialize16(motor       [i]);     // 32-47: motor
  for (i = 0; i < 8; i++) deserialize16(rcHysteresis[i]);     // 48-63: receiver

  int8_t sensorData;  
  deserialize8(sensorData);
        
//    nunchukPres = (sensorData & 1) != 0;
  accPres     = (sensorData & 3);              // 64: avail. sensors:   0-nunchukPres, 1-accPres, 2-baroPres, 3-magPres
  baroPres    = (sensorData & 4);
  magPres     = (sensorData & 8);

  deserialize8(sensorData);                    // 65: sensor modes:     0-accMode, 1-baroMode, 2-magMode
  accMode     = (sensorData & 1);
  baroMode    = (sensorData & 2);
  magMode     = (sensorData & 4);

  deserialize16(cycleTime);                    // 66-67: cycle time
  
  for (i = 0; i < 2; i++) {
    deserialize16(angle[i]);                   // 68-71: roll and pitch angle
  }
  
#ifdef TEST
  bptr++;
#else
  deserialize8(config);                        // 72: config / multitype (Tri, Quad, ...)
#endif

  for (i = 0; i < 5; i++) {
    deserialize8(P8[i]);                       // 73-87: P-I-D (roll, pitch, yaw, alt, vel)
    deserialize8(I8[i]);
    deserialize8(D8[i]);
  }
  deserialize8(P8[5]); deserialize8(I8[5]);    // 88-89: pLevel, iLevel
  deserialize8(P8[6]);                         // 90:    pMag
  
  deserialize8(rcRate8);                       // 91: rcRate (dual rate)
  deserialize8(rcExpo8);                       // 92: rcExpo
  deserialize8(rollPitchRate);                 // 93: roll/pitch rate
  deserialize8(yawRate);                       // 94: yaw rate
  deserialize8(dynThrPID);                     // 95: dyn. throttle PID
    
  for (i = 0; i < 6; i++) {                    // 96-101: activation modes (level, baro, mag, camStab, camTrig, arming)
    deserialize8(actMode[i]);
  }

  deserialize16(powerMeterSum);                // 102-103: powerMeterSum
  deserialize16(powerTrigger);                 // 104-105: powerTrigger
  serialize8(vbat);                            // 106: vbat (battery level)

  uint16_t temp;
  deserialize16(temp);                         // 107-114: 4 16bit debug values
  deserialize16(temp);                                  
  deserialize16(temp);                                   
  deserialize16(temp);                                  
}

void serialWrite() {
  lcdNoBlink();
  
  bptr = buffer;
  int8_t i;
  
  serialize8('W');                        // 00: tag
  for (i = 0; i < 5; i++) {               // 01-15: PID (roll, pitch, yaw, alt, vel)
    serialize8(P8[i]); 
    serialize8(I8[i]); 
    serialize8(D8[i]);
  }
  serialize8(P8[5]); serialize8(I8[5]);   // 16-17: pLevel, iLevel
  serialize8(P8[6]);                      // 18: P mag

  serialize8(rcRate8);                      // 19: rcRate (dual rate)
  serialize8(rcExpo8);                    // 20: rcExpo
  serialize8(rollPitchRate);              // 21: rollPitchRate
  serialize8(yawRate);                    // 22: yawRate
  serialize8(dynThrPID);                  // 23: dyn. Throttle PID

  for (i = 0; i < 6; i++) {               // 24-29: activation modes (level, baro, mag, camStab, camTrig, arming)
    serialize8(actMode[i]);
  }

  serialize16(powerTrigger);              // 30-31: powerTrigger
  
  serialFlush();
//  while (serialAvailable()) {
//    serialReadByte();
//    flashRed();
//  }
  
  uint8_t len = bptr - buffer;
//  serialWriteBuffer();
  Serial.write(buffer, len);
  serialFlush();

  lcdPrintProg1(writeMsg);
  format(7 + 16, len, 2, 0);

  blinkConfirm();

  serialRead(); // reload the frame
}


/* ----- Persistence: ----- */

uint8_t readEEPROM ()              {return EEPROM.read(ee_ptr++);}
void    writeEEPROM(uint8_t val)   {if (EEPROM.read(ee_ptr) != val) {EEPROM.write(ee_ptr, val);} ee_ptr++;}

void storeProfile() {
  getProfileName(lineBuffer); 
  if (enterName(lineBuffer)) {
    storeProfile(lineBuffer);
  }
}

boolean restoreProfile() {
  ee_ptr = profile * 48 + 16 + 8; // skip name

  uint8_t* lptr = buffer;
  for (int8_t i = 0; i < 32; i++) {
    *lptr++ = readEEPROM();
  }
  return extractProfile(buffer);
}
 
boolean extractProfile(uint8_t* lptr) { 
  uint8_t tag = *lptr++;                    // 00: version

  int8_t i;
  if (tag == 18) {
    for (i = 0; i < 5; i++) {               // 01-15: PID (roll, pitch, yaw, alt, vel)
      P8[i] = *lptr++; 
      I8[i] = *lptr++; 
      D8[i] = *lptr++;
    }
    P8[5] = *lptr++; I8[5] = *lptr++;       // 16-17: pLevel, iLevel
    P8[6] = *lptr++;                        // 18: P mag

    rcRate8       = *lptr++;                // 19: rcRate (dual rate)
    rcExpo8       = *lptr++;                // 20: rcExpo
    rollPitchRate = *lptr++;                // 21: rollPitchRate
    yawRate       = *lptr++;                // 22: yawRate
    dynThrPID     = *lptr++;                // 23: dyn. Throttle PID

    for (i = 0; i < 6; i++) {               // 24-29: activation modes (level, baro, mag, camStab, camTrig, arming)
      actMode[i] = *lptr++;
    }

    uint8_t high = *lptr++;
    uint8_t low  = *lptr++;
    powerTrigger = high * 256 + low;        // 30-31: powerTrigger
    
    serialWrite();
    return true;
  }
  return false;
}

void storeProfile(char* name) {
  uint8_t* lptr = buffer;

  int8_t i = 0;
  while (name[i] != '\0' && i < 8) { *lptr++ = name[i++]; }
  while (i < 8)                    { *lptr++ = ' ';       }
  
  *lptr++ = vers;                         // 00: version
  for (i = 0; i < 5; i++) {               // 01-15: PID (roll, pitch, yaw, alt, vel)
    *lptr++ = P8[i];
    *lptr++ = I8[i]; 
    *lptr++ = D8[i];
  }
  *lptr++ = P8[5]; *lptr++ = I8[5];       // 16-17: pLevel, iLevel
  *lptr++ = P8[6];                        // 18: P mag

  *lptr++ = rcRate8;                      // 19: rcRate (dual rate)
  *lptr++ = rcExpo8;                      // 20: rcExpo
  *lptr++ = rollPitchRate;                // 21: rollPitchRate
  *lptr++ = yawRate;                      // 22: yawRate
  *lptr++ = dynThrPID;                    // 23: dyn. Throttle PID

  for (i = 0; i < 6; i++) {               // 24-29: activation modes (level, baro, mag, camStab, camTrig, arming)
    *lptr++ = actMode[i];
  }
  *lptr++ = powerTrigger / 256;           // 30-31: powerTrigger
  *lptr++ = powerTrigger % 256;
  
  lptr   = buffer;
  ee_ptr = profile * 48 + 16;
  for (i = 0; i < 40; i++) {
    writeEEPROM(*lptr++);
  }
}

void deleteProfile() {
  ee_ptr = profile * 48 + 16;
  for (int8_t i = 0; i < 40; i++) {
    writeEEPROM(0);
  }
}

bool getProfileName(char* name) {
  ee_ptr = profile * 48 + 16;
  for (int8_t i = 0; i < 8; i++) { name[i] = readEEPROM(); }
  //name[8] = '\0'; 

  uint8_t tag  = readEEPROM();
  boolean succ = (tag == 18);  
  if (!succ) {
    for (int8_t i = 0; i < 8; i++) { name[i] = '-'; }
  }
  return succ;
} 

// ----------------

#ifndef STANDALONE  
void restoreParams() {
  ee_ptr   = 0;
  baudRate = DEFAULT_BAUDRATE;    // default value: 8 = 115200 Baud

  byte magic0 = readEEPROM();     // check validity of data by storing some magic numbers as a prefix
  byte magic1 = readEEPROM();
  ee_ptr++;                       // was: MWC version

  if (magic0 == 47 && magic1 == 11) {
#ifdef LOGGING
    int8_t tmp = readEEPROM();
    if (tmp >= NO_LOGGING) {
      logMode  = tmp;
    }
#else
  ee_ptr++;
#endif
    
    delayTime =       readEEPROM();
    backlight = 255 - readEEPROM();
    
#ifndef USE_DEFAULT_BAUDRATE    
    {
      int8_t tmp = -1 + readEEPROM();
      if (tmp < 8) {
        baudRate = tmp;
      }
    }
#endif    
  }
}

void resetLogMode() {
#ifdef LOGGING
  ee_ptr = 3;
  writeEEPROM(NO_LOGGING);
#endif  
}

void storeParams() {
  ee_ptr = 0;
  writeEEPROM(47);  
  writeEEPROM(11);
  ee_ptr++;          // was: MWC version
  writeEEPROM(logMode);
  writeEEPROM(delayTime);
  writeEEPROM(255 - backlight);
  writeEEPROM(baudRate + 1);
}

#else // STANDALONE  
void restoreParams() {
  logMode   = LF_LOGVIEW;   // LogView format
  delayTime = 25;           // sample rate in 10ms (25 == 250 ms == 4 samples per second)
  baudRate  = 0;            // has to match with SERIAL_COM_SPEED in MWC
                            //  0 =   1200 bps,  1 =   2400 bps,  2 =   4800 bps,  3 =   9600 bps
                            //  4 =  14400 bps,  5 =  19200 bps,  6 =  38400 bps,  7 =  57600 bps
                            //  8 = 115200 bps  
}

void resetLogMode() {
}

void storeParams() {
}
#endif // STANDALONE  


/* ----- Logging: ----- */

#ifdef LOGGING
boolean fileExists() { return SD.exists(fileName); }

void showLineBlink(prog_char const line[]) {
  lcdPrintProg1(line);
  blinkWarning();
}

void openLog() {
  int cnt    = 0;
  int offset = 0;
#ifdef LOGVIEW
  if (logMode == LF_LOGVIEW) {
    offset = 500;
  }
#endif

  //  boolean created = false;
  while (!logFile && cnt < 500) {
    formatLogFile(offset + cnt);
    if (!fileExists()) {
      logFile = SD.open(fileName, FILE_WRITE);
    }
    else {
      cnt++;
    }
  }
  if (!logFile) {
    resetLogMode();
    setError(fileError);
  }
//  state = 0;
  storeParams();

  lcdPrintProg0(logMsg);
//  printFileName();

  if (logMode != LF_LOGVIEW) {
    logFile.write(vers);
    logFile.write(config);
  }

  setBacklight();
  green(true);
}

#ifdef LOGVIEW
byte* toBuffer(byte* outPtr, int16_t value[3]) {return toBuffer(outPtr, value[0]);}

byte* toBuffer(byte* outPtr, int16_t value   ) {
  if (value == 0) {
    *outPtr++ = '0';
  }
  else {
    if (value < 0) {
      value = -value;
      *outPtr++ = '-';
    }
    boolean flag = false;
    int16_t dec  = 10000;
    while (dec > 0) {
      int val = (value / dec) % 10;
      flag |= (val > 0);
      if (flag) {
        *outPtr++ = val + '0';
      }
      dec /= 10;
    }
  }
  *outPtr++ = ';';
  return outPtr;
}

byte* toBufferPwm(byte* outPtr, int16_t value[3], boolean sym) {return toBufferPwm(outPtr, value[0], sym);}

byte* toBufferPwm(byte* outPtr, int16_t value   , boolean sym) {
  if (value != 0) {
    if (sym) {
      value = (value - 1500) / 5; // symmetric [-100..+100]
    }
    else {
      value = (value - 1000) / 10; // percent (0..100)
    }
    outPtr = toBuffer(outPtr, value);
  }
  else {
    *outPtr++ = ';';
  } 
  return outPtr;
}

byte* addPrefix(byte* outPtr, char channel) {
  *outPtr++ = '$';
  *outPtr++ = channel;
  *outPtr++ = ';';
  *outPtr++ = '1' + accMode + baroMode + magMode;  // state
  *outPtr++ = ';';
  outPtr = toBuffer(outPtr, lineNo); // used as timeStamp
  return outPtr;
}

byte* addSuffix(byte* outPtr) {
  *outPtr++ = '0';  // checksum - not used
  *outPtr++ = '\r';
  *outPtr++ = '\n';
  return outPtr;
}
#endif

//int8_t channels[] = { THROTTLE, ROLL, PITCH, YAW, AUX1, AUX2, CAM_ROLL, CAM_PITCH, };
#ifdef LOGVIEW
static uint8_t channelMapping[] = { 3, 0, 1, 2, 4, 5, 6, 7, };
#endif

void doLog() {
  state  = LOG_STATE;
  lineNo = 0;

  while (logMode > NO_LOGGING) {
#ifdef LOGVIEW
    if (logMode == LF_LOGVIEW) {
      deserialize();
  
      byte  outBuffer[160];
      byte* outPtr = outBuffer;
  
      outPtr = addPrefix(outPtr, '1');
      uint8_t i;
      for (i = 0; i < 3; i++) { outPtr = toBuffer(outPtr, accSmooth[i]); } // 2 - 4
      for (i = 0; i < 3; i++) { outPtr = toBuffer(outPtr, gyroData [i]); } // 5 - 7
      for (i = 0; i < 3; i++) { outPtr = toBuffer(outPtr, magADC   [i]); } // 8 - 10
      outPtr = toBuffer(outPtr, altitude);                                 // 11
      outPtr = toBuffer(outPtr, heading );                                 // 12
      for (i = 0; i < 2; i++) { outPtr = toBuffer(outPtr, angle[i]); }     // 13 - 14
      outPtr = addSuffix(outPtr);
  
      outPtr = addPrefix(outPtr, '2');
      for (i = 0; i < 4; i++) { outPtr = toBufferPwm(outPtr, servo[i], true);  } // 1 - 4
      for (i = 0; i < 8; i++) { outPtr = toBufferPwm(outPtr, motor[i], false); } // 5 - 12
      outPtr = addSuffix(outPtr);
  
      outPtr = addPrefix(outPtr, '3');
      for (i = 0; i < 8; i++) { outPtr = toBufferPwm(outPtr, rcHysteresis[channelMapping[i]], i); } // 1 - 8 --- last arg false for i=0, true otherwise
      outPtr = addSuffix(outPtr);
      
      logFile.write(outBuffer, outPtr - outBuffer);
    }
    else
#endif
    {
      logFile.write(buffer, bptr - buffer);
    }
    setBacklight();
  
    lineNo++;
    
    int8_t key = getKeyPressed();
    if (key >= 0) {
      bool flag = confirm(stopLogMsg);
  
      if (flag) {
        logMode = NO_LOGGING;
        break;
      }
      else {
        lcdPrintProg0(logMsg);
//        printFileName();
      }
    }
    if (serialRead() > 0) {
      printFileName();
      format(11 + 16, lineNo, 5, 0);
      flashYellow();
      delayT();
    }
    else {
      setOffline(timeoutError);
    }
  }
  resetLogMode();
  closeLogFile();
}

int findNextLogFile(int act, boolean ascend, boolean binOnly) {
  int     mod   = (binOnly)? 500 : 1000;
  int     offs  = (ascend)? 1 : mod-1;
  int     cnt   = (act + offs) % mod;
  boolean found = false;
  while (!found && cnt != act) {
    formatLogFile(cnt);
    red(!(cnt & 0x1F));
    found = fileExists();
    if (!found) {
      cnt = (cnt + offs) % mod;
    }
  }
  red(false);
  if (!found) {
    formatLogFile(act);          // check if exactly one logfile exists
    found = fileExists();
  }
  setBacklight();
  return (!found)? -1 : cnt;
}

void formatLogFile(int cnt) {
  char c5 = 'o', c6 = 'g'; // .log - binary
  if (cnt >= 500) {
    c5 = 'v';              // .lvw - logview
    c6 = 'w';
  }
  fileName[5] = c5;
  fileName[6] = c6;

  fileName[2] = '0' +  cnt     % 10;
  fileName[1] = '0' + (cnt/10) % 10;
  fileName[0] = '0' + (cnt/100);
}

void printFileName() {
  lcdPrintProg1(fileMsg);
  for (int8_t i = 0; i < 3; i++) {
    lcdWrite(fileName[i]);
  }
  lcdPrint(" #");
}

uint8_t readFromFile() {return logFile.read();}

int selectLogfile(int next, boolean binOnly) {
  boolean cancel = false;
  next = findNextLogFile(next, true, binOnly);
  while (next >= 0) {
    boolean valid = false;
    
    uint8_t confId = -1;
    uint8_t fileVers;
    prog_char const* typeStr = configIds;
    if (next < 500) {
      if (openFileRead()) {
        fileVers = readFromFile();
        confId   = readFromFile();
        valid    = (fileVers == 18);
        if (valid) {
          typeStr = versions[fileVers - 18];
        }
        closeLogFile();
      }
    }
    else {
      valid   = true;
      typeStr = emptyLine;
    }
    
    printFileName();
    lcdPrintProg(10 + 16, typeStr);
    if (confId >= 0 && confId < NUM_CONFIGS) {
      char* chptr = shortConfigIds + (confId*2);
      lcdWrite(*chptr++);
      lcdWrite(*chptr++);
    }

    int8_t key = enterKey();
    if (key == KEY_UP || key == KEY_DOWN) {
      next = findNextLogFile(next, key == KEY_DOWN, binOnly);
    }
    else if (key == KEY_LEFT) {
      next   = -1;
      cancel = true;
    }
    else if (key >= 0 && valid) { 
      break;
    }
  }
  return next;
}

boolean openFileRead() {
  logFile = SD.open(fileName, FILE_READ);

  boolean succ = logFile;
  if (!succ) {
    showLineBlink(fileError);
  }
  green(succ);
  return succ;
}

void closeLogFile() {
  logFile.close();
  setBacklight();
  green(false);
}

void playbackLog() {
  enableAPC220(false);
  Serial.begin(GUI_SERIAL_COM_SPEED);

  int fileNo = 500;
  while (fileNo >= 0) {
    fileNo = selectLogfile(fileNo-1, true);
    if (fileNo >= 0) {
      lineNo = 0;
      if (openFileRead()) {
        uint8_t fileVers = readFromFile() - 18;
        uint8_t confId   = readFromFile();
        if (fileVers < sizeof(frameLength)) {
          uint8_t len = frameLength[fileVers] - 1;
          while (logFile.available()) {
            if (getKeyPressed() == KEY_LEFT) {
              showLineBlink(cancelRead);
              break;
            }
            
            printFileName();
            format(11 + 16, lineNo, 5, 0);
            lineNo++;
          
            bptr = buffer;
            uint8_t readBytes = 0;
            while (readBytes < len && logFile.available()) {
              *bptr++ = readFromFile();
              readBytes++;
            }
            setBacklight();
            if (readBytes == len) {
              int     timeout = 0;
              boolean cancel  = false;
              char    next    = '\0';
              
              while (!cancel && timeout < 600 && next != 'M') {
                if (serialAvailable()) {
                  next = serialReadByte();
                }
                if (!(timeout & 15)) {
                  flashRed();
                }
                if (getKeyPressed() == KEY_LEFT) {
                  cancel = true;
                }
                delay50();
                timeout++;
              }
              if (cancel || timeout >= 600) {
                showLineBlink(cancel? cancelRead : timeoutError);
                break;
              }                
              flashYellow();
              *bptr++ = 'M';
              Serial.write(buffer, bptr - buffer);
  //              serialFlush();
              delayT(); // in real-time
            }
            else {
              break; // eof
            }
          }
        } // wrong version
      }
      closeLogFile();
      lcdPrintProg1(emptyLine);
    }
  }
  enableAPC220(true);
  delay50();
  serialBegin();
}
#endif // LOGGING

// --ar 13.08.2011

